from __future__ import annotations

import os
import re
from typing import (
    AnyStr,
    Callable,
    Generic,
    List,
    Optional,
    Pattern,
    Tuple,
    Union,
)

from e3.diff import diff
from e3.os.fs import unixpath
from e3.testsuite.result import FailureReason, Log, binary_repr, truncated
from e3.testsuite.driver.classic import ClassicTestDriver, TestAbortWithError
from e3.testsuite.utils import indent


class OutputRefiner(Generic[AnyStr]):
    """
    Interface to refine a test output before baseline and actual comparison.

    Sometimes, the way a library/tool works forces it to have outputs that
    depends on the environment (for instance: the location of the testsuite on
    the filesystem, the current date, etc.). Refiners makes it possible for
    a testsuite to "hide" these discrepancies during the diff computation.

    Note that output refiners might get called bytes strings when the test
    drivers operate in binary mode.
    """

    def refine(self, output: AnyStr) -> AnyStr:
        """
        Refine a test/baseline output.

        :param output: Output to refine.
        """
        raise NotImplementedError


class RefiningChain(OutputRefiner[AnyStr]):
    """Simple wrapper for a sequence of output refiners applied in chain."""

    def __init__(self, refiners: List[OutputRefiner]):
        """
        Initialize a RefiningChain instance.

        :param refiners: List of refiners to execute in sequence.
        """
        self.refiners = refiners

    def refine(self, output: AnyStr) -> AnyStr:
        for r in self.refiners:
            output = r.refine(output)
        return output


class Substitute(OutputRefiner[AnyStr]):
    """Replace substrings in outputs."""

    def __init__(
        self, substring: AnyStr, replacement: Optional[AnyStr] = None
    ) -> None:
        """
        Initialize a Substitute instance.

        :param substring: Substring to replace.
        :param replacement: Replacement to use for the substitution. If left to
            None, just remove the substring (i.e. use an empty replacement
            string).
        """
        self.substring: AnyStr = substring
        self.replacement: AnyStr = replacement or type(substring)()

    def refine(self, output: AnyStr) -> AnyStr:
        return output.replace(self.substring, self.replacement)


class CanonicalizeLineEndings(OutputRefiner[AnyStr]):
    r"""Replace \r\n with \n in outputs."""

    def refine(self, output: AnyStr) -> AnyStr:
        if isinstance(output, str):
            return output.replace("\r\n", "\n")
        else:
            return output.replace(b"\r\n", b"\n")


class ReplacePath(RefiningChain[str]):
    """Return an output refiner to replace the given path."""

    def __init__(self, path: str, replacement: str = "") -> None:
        # First replace the normalized path, then the Unix-style path (which
        # some tool may output even on Windows) and finally the very path that
        # was given.
        super().__init__(
            [
                Substitute(substring, replacement)
                for substring in [os.path.realpath(path), unixpath(path), path]
            ]
        )


class PatternSubstitute(OutputRefiner, Generic[AnyStr]):
    """Replace patterns in outputs."""

    def __init__(
        self,
        pattern: AnyStr,
        replacement: AnyStr | Callable[[re.Match], AnyStr] | None = None,
    ) -> None:
        """
        Initialize a PatternSubstitute instance.

        :param pattern: Pattern (regular expression) to replace.
        :param replacement: Replacement to use for the substitution. If left to
            None, just remove the substring (i.e. use an empty replacement
            string).
        """
        self.regexp: Pattern[AnyStr] = re.compile(pattern)
        self.replacement: AnyStr | Callable[[re.Match], AnyStr] = (
            replacement or type(pattern)()
        )

    def refine(self, output: AnyStr) -> AnyStr:
        return self.regexp.sub(self.replacement, output)


class LineByLine(OutputRefiner[AnyStr]):
    """Wrapper to apply an output refine line by line."""

    def __init__(self, refiner: OutputRefiner) -> None:
        """
        Initialize a LineByLine instance.

        :param refiner: Refiner to apply on each input line.
        """
        self.refiner = refiner

    def refine(self, output: AnyStr) -> AnyStr:
        if isinstance(output, str):
            return "".join(
                self.refiner.refine(line)
                for line in output.splitlines(keepends=True)
            )
        else:
            return b"".join(
                self.refiner.refine(line)
                for line in output.splitlines(keepends=True)
            )


class DiffTestDriver(ClassicTestDriver):
    """Test driver to compute test output against a baseline."""

    @property
    def rewrite_baseline(self) -> bool:
        """Return whether this driver should rewrite its baseline."""
        # If a failure is expected for this test, consider that the actual
        # output is wrong, and so do not rewrite the baseline.
        return (
            self.baseline_file is not None
            and not self.test_control.xfail
            and getattr(self.env, "rewrite_baselines", False)
        )

    @property
    def baseline_file(self) -> Tuple[str, bool]:
        """Return the test output baseline file.

        :return: The name of the text file (relative to test directories) that
            contains the expected test output and whether the baseline is a
            regexp.
        """
        filename = self.test_env.get("baseline_file", "test.out")
        is_regexp = self.test_env.get("baseline_regexp", False)
        return (filename, is_regexp)

    @property
    def baseline(self) -> Tuple[Optional[str], Union[str, bytes], bool]:
        """Return the test output baseline.

        Subclasses can override this method if they want to provide a baseline
        that does not come from a file, short-circuiting the baseline_file
        property.

        :return: The baseline absolute filename (if any), the baseline content,
            as a string or as a bytes string (depending on the default
            encoding), and whether the baseline is a regexp. The baseline
            filename is used to rewrite test output: leave it to None if
            rewriting does not make sense.
        """
        encoding = self.default_encoding
        is_binary = encoding == "binary"
        filename, is_regexp = self.baseline_file
        filename = self.test_dir(filename)
        baseline: Union[str, bytes]

        # If the baseline should be rewritten, tolerate a missing baseline file
        # and start from an empty baseline: we will create the baseline once
        # the test has run.
        if self.rewrite_baseline and not os.path.isfile(filename):
            return (filename, b"" if is_binary else "", is_regexp)

        try:
            if is_binary:
                with open(filename, "rb") as bin_f:
                    baseline = bin_f.read()
            else:
                with open(filename, "r", encoding=encoding) as text_f:
                    baseline = text_f.read()
        except Exception as exc:
            raise TestAbortWithError(
                "cannot read baseline file ({}: {})".format(
                    type(exc).__name__, exc
                )
            ) from exc
        return (filename, baseline, is_regexp)

    @property
    def output_refiners(self) -> List[OutputRefiner]:
        """
        List of refiners for test baselines/outputs.

        This just returns a refiner to canonicalize line endings unless the
        test environment contains a "strict_line_endings" key associated to
        true.
        """
        return (
            []
            if self.test_env.get("strict_line_endings", False)
            else [CanonicalizeLineEndings()]
        )

    @property
    def refine_baseline(self) -> bool:
        """Whether to apply output refiners to the output baseline."""
        return True

    @property
    def diff_ignore_white_chars(self) -> bool:
        """
        Whether to ignore white characters in diff computations.

        This returns whether the comparison between test output and baseline
        must ignore whitespaces (leading and trailing spaces, tabs and carriage
        returns on lines, and empty lines). Note that if we don't ignore them,
        we still canonicalize line separators (CRLF are replaced by LF before
        the comparison).

        Note that at some point, this mechanism should be unified with the
        ``output_refiners`` machinery. However, this relies on e3.diff's
        ignore_white_chars feature, which is not trivial to reimplement.
        """
        return False

    @property
    def diff_context_size(self) -> int:
        """Positive number of context lines to include in diff computations."""
        return self.test_env.get("diff_context_size", 1)

    def set_up(self) -> None:
        super().set_up()

        # Keep track of the number of non-clean diffs
        self.failing_diff_count = 0

    def compute_diff(
        self,
        baseline_file: Optional[str],
        baseline: AnyStr,
        actual: AnyStr,
        failure_message: str = "unexpected output",
        ignore_white_chars: Optional[bool] = None,
        context_size: Optional[int] = None,
        truncate_logs_threshold: Optional[int] = None,
    ) -> List[str]:
        """Compute the diff between expected and actual outputs.

        Return an empty list if there is no diff, and return a list that
        contains an error message based on ``failure_message`` otherwise.

        :param baseline_file: Absolute filename for the text file that contains
            the expected content (for baseline rewriting, if enabled), or None.
        :param actual: Actual content to compare.
        :param failure_message: Failure message to return if there is a
            difference.
        :param ignore_white_chars: Whether to ignore whitespaces during the
            diff computation. If left to None, use
            ``self.diff_ignore_white_chars``.
        :param context_size: Positive number of context lines to include in
            diff computations. If left to None, use ``self.diff_context_size``.
        :param truncate_logs_threshold: Threshold to truncate the diff message
            in ``self.result.log``. See ``e3.testsuite.result.truncated``'s
            ``line_count`` argument. If left to None, use the testsuite's
            ``--truncate-logs`` option.
        """
        if ignore_white_chars is None:
            ignore_white_chars = self.diff_ignore_white_chars

        if context_size is None:
            context_size = self.diff_context_size

        if truncate_logs_threshold is None:
            truncate_logs_threshold = self.testsuite_options.truncate_logs

        # Run output refiners on the actual output, not on the baseline
        refiners = (
            RefiningChain[str](self.output_refiners)
            if isinstance(actual, str)
            else RefiningChain[bytes](self.output_refiners)
        )
        refined_actual = refiners.refine(actual)
        refined_baseline = (
            refiners.refine(baseline) if self.refine_baseline else baseline
        )

        # When running in binary mode, make sure the diff runs on text strings
        if self.default_encoding == "binary":
            assert isinstance(refined_actual, bytes)
            assert isinstance(refined_baseline, bytes)
            decoded_actual = binary_repr(refined_actual)
            decoded_baseline = binary_repr(refined_baseline)
        else:
            assert isinstance(refined_actual, str)
            assert isinstance(refined_baseline, str)
            decoded_actual = refined_actual
            decoded_baseline = refined_baseline

        # Get the two texts to compare as list of lines, with trailing
        # characters preserved (splitlines(keepends=True)).
        expected_lines = decoded_baseline.splitlines(True)
        actual_lines = decoded_actual.splitlines(True)

        # Compute the diff. If it is empty, return no failure. Otherwise,
        # include the diff in the test log and return the given failure
        # message.
        d = diff(
            expected_lines,
            actual_lines,
            ignore_white_chars=ignore_white_chars,
            context=context_size,
        )
        if not d:
            return []

        self.failing_diff_count += 1
        message = failure_message

        diff_lines = []
        for line in d.splitlines():
            # Add colors diff lines
            if line.startswith("-"):
                color = self.Fore.RED
            elif line.startswith("+"):
                color = self.Fore.GREEN
            elif line.startswith("@"):
                color = self.Fore.CYAN
            else:
                color = ""
            diff_lines.append(color + line + self.Style.RESET_ALL)

        # If requested, rewrite the test baseline with the actual output
        if self.rewrite_baseline:
            assert baseline_file is not None
            if isinstance(refined_actual, str):
                with open(
                    baseline_file, "w", encoding=self.default_encoding
                ) as f:
                    f.write(refined_actual)
            else:
                assert isinstance(refined_actual, bytes)
                with open(baseline_file, "wb") as f:
                    f.write(refined_actual)
            message = "{} (baseline updated)".format(message)

        # Send the appropriate logging. Make sure ".log" has all the
        # information. If there are multiple diff failures for this testcase,
        # do not emit the "expected/out" logs, as they support only one diff.
        diff_log = (
            self.Style.RESET_ALL
            + self.Style.BRIGHT
            + "Diff failure: {}\n".format(message)
            + "\n".join(diff_lines)
            + "\n"
        )
        self.result.log += "\n" + truncated(diff_log, truncate_logs_threshold)
        if self.failing_diff_count == 1:
            self.result.expected = Log(decoded_baseline)
            self.result.out = Log(decoded_actual)
            self.result.diff = Log(diff_log)
        else:
            self.result.expected = None
            self.result.out = None
            assert isinstance(self.result.diff, Log) and isinstance(
                self.result.diff.log, str
            )
            self.result.diff += "\n" + diff_log

        return [message]

    def compute_regexp_match(
        self,
        regexp: Union[Pattern[AnyStr], AnyStr],
        actual: AnyStr,
        failure_message: str = "output does not match expected pattern",
        truncate_logs_threshold: Optional[int] = None,
    ) -> List[str]:
        """Compute whether the actual output matches a regexp.

        Return an empty list if the acutal content matches, and return a list
        that contains an error message based on ``failure_message`` otherwise.

        :param regexp: Regular expression to use.
        :param actual: Actual content to match.
        :param failure_message: Failure message to return if there is a
            difference.
        :param truncate_logs_threshold: Threshold to truncate the diff message
            in ``self.result.log``. See ``e3.testsuite.result.truncated``'s
            ``line_count`` argument. If left to None, use the testsuite's
            ``--truncate-logs`` option.
        """
        if isinstance(regexp, (str, bytes)):
            regexp = re.compile(regexp)

        if truncate_logs_threshold is None:
            truncate_logs_threshold = self.testsuite_options.truncate_logs

        # Run output refiners. Code is more complex than it should be to
        # satisfy Mypy's constraints.
        refiners = (
            RefiningChain[str](self.output_refiners)
            if isinstance(actual, str)
            else RefiningChain[bytes](self.output_refiners)
        )
        refined_actual = refiners.refine(actual)

        match = regexp.fullmatch(refined_actual)
        if match:
            return []

        def quote(content: AnyStr) -> str:
            decoded_content: str
            if self.default_encoding == "binary":
                assert isinstance(content, bytes)
                decoded_content = binary_repr(content)
            else:
                assert isinstance(content, str)
                decoded_content = content
            return indent(decoded_content)

        # Send the appropriate logging
        self.result.log += failure_message + ":\n"
        self.result.log += truncated(
            quote(refined_actual), truncate_logs_threshold
        )
        self.result.log += "\nDoes not match the expected pattern:\n"
        self.result.log += truncated(
            quote(regexp.pattern), truncate_logs_threshold
        )

        return [failure_message]

    def compute_failures(self) -> List[str]:
        """Return a failure if ``self.output.log`` does not match the baseline.

        This computes a diff with the content of the baseline file, unless
        there is a "baseline_regexp" entry in the test environment that
        evaluates to true.

        Subclasses can override this if they need more involved analysis:
        for instance computing multiple diffs.
        """
        # Get the baseline, then check that the actual output matches
        filename, baseline, is_regexp = self.baseline
        if is_regexp:
            result = self.compute_regexp_match(baseline, self.output.log)
        else:
            # Run output refiners. Code is more complex than it should be to
            # satisfy Mypy's constraints.
            if isinstance(baseline, str):
                assert isinstance(self.output.log, str)
                result = self.compute_diff(filename, baseline, self.output.log)
            else:
                assert isinstance(baseline, bytes)
                assert isinstance(self.output.log, bytes)
                result = self.compute_diff(filename, baseline, self.output.log)

        # Adjust the result if there is a mismatch
        if result:
            self.result.failure_reasons.add(FailureReason.DIFF)

        return result



from __future__ import annotations

"""
RunningStatus keeps users informed about the progress of testsuite execution
and aborts when there are too many consecutive failures/errors (see the
--max-consecutive-failures command-line option).
"""

from typing import Dict, Optional, TYPE_CHECKING
import threading
import time

from e3.testsuite.result import TestResultSummary, TestStatus


if TYPE_CHECKING:
    from e3.collection.dag import DAG
    from e3.testsuite.fragment import TestFragment


class RunningStatus:
    def __init__(
        self,
        filename: str,
        update_interval: float = 1.0,
        max_consecutive_failures: int = 0,
    ):
        """RunningStatus constructor.

        :param filename: Name of the status file to write.
        :param update_interval: Minimum number of seconds between status file
            updates.
        :param max_consecutive_failures:
            Number of test failures (FAIL or ERROR) that trigger the abortion
            of the testuite. If zero, this behavior is disabled.
        """
        self.dag: Optional[DAG] = None
        self.filename = filename
        self.lock = threading.Lock()

        self.running: Dict[str, TestFragment] = {}
        """Set of test fragments currently running, indexed by UID."""

        self.completed: Dict[str, TestFragment] = {}
        """Set of test fragments that completed their job, indexed by UID."""

        self.status_counters: Dict[TestStatus, int] = {}
        """Snapshot of the testsuite's report index status counters.

        We preserve a copy to avoid inconsistent state due to race conditions:
        these counters are updated in the collect_result method while the other
        sets are updated in TestFragment.run, which is executed in workers
        (i.e. other threads).
        """

        self.update_interval = update_interval
        self.no_update_before = 0.0

        self.max_consecutive_failures = max_consecutive_failures

        self.consecutive_failures = 0
        """
        Number of consecutive failure/error results we just processed. Used to
        abort the testsuite when there are too many issues.
        """

        self.aborted_too_many_failures = False
        """
        Whether the testsuite aborted because of too many consecutive test
        failures (see the --max-consecutive-failures command-line option).
        """

    def set_dag(self, dag: DAG) -> None:
        """Set the DAG that contains TestFragment instances."""
        assert self.dag is None
        self.dag = dag

    def start(self, fragment: TestFragment) -> None:
        """Put a fragment in the "running" set."""
        assert self.dag is not None
        with self.lock:
            assert fragment.uid in self.dag.vertex_data
            assert fragment.uid not in self.running
            assert fragment.uid not in self.completed
            self.running[fragment.uid] = fragment
        self.dump()

    def complete(self, fragment: TestFragment) -> None:
        """Move a fragment from the "running" set to the "completed" set."""
        assert self.dag is not None
        with self.lock:
            assert fragment.uid in self.dag.vertex_data
            assert fragment.uid not in self.completed
            f = self.running.pop(fragment.uid)
            assert f is fragment
            self.completed[f.uid] = f

        self.dump()

    def process_result(self, result: TestResultSummary) -> None:
        """Integrate a test result in the testsuite status.

        This increments status counters and triggers testsuite abortion if
        there were too many consecutive failures.
        """
        with self.lock:
            try:
                self.status_counters[result.status] += 1
            except KeyError:
                self.status_counters[result.status] = 1

            # Keep track of the number of consecutive failures seen so far: if
            # it reaches the maximum number allowed, we must abort the
            # testsuite.
            if result.status in (TestStatus.ERROR, TestStatus.FAIL):
                self.consecutive_failures += 1
                if (
                    not self.aborted_too_many_failures
                    and self.max_consecutive_failures > 0
                    and (
                        self.consecutive_failures
                        >= self.max_consecutive_failures
                    )
                ):
                    from e3.testsuite import logger

                    self.aborted_too_many_failures = True
                    logger.error(
                        "Too many consecutive failures, aborting the testsuite"
                    )
            else:
                self.consecutive_failures = 0

        self.dump()

    def dump(self) -> None:
        """Write a report for this status as human-readable text to "fp"."""
        # Do not update the status file more than once per second
        now = time.time()
        if self.update_interval and now < self.no_update_before:
            return
        self.no_update_before = now + self.update_interval

        lines = []
        with self.lock:
            if self.dag is None:
                lines.append("No test fragment yet")
            else:
                lines.append(
                    f"Test fragments:"
                    f" {len(self.completed)}"
                    f" / {len(self.dag.vertex_data)} completed"
                )

                lines.append("Currently running:")
                if self.running:
                    lines.extend(f"  {uid}" for uid in sorted(self.running))
                else:
                    lines.append("  <none>")

                statuses = sorted(
                    [
                        (status, count)
                        for status, count in self.status_counters.items()
                        if count
                    ],
                    key=lambda couple: couple[0].value,
                )
                lines.append("Partial results:")
                if statuses:
                    for status, count in statuses:
                        lines.append(f"  {status.name.ljust(12)} {count}")
                else:
                    lines.append("  <none>")

        text = "\n".join(lines)
        with open(self.filename, "w") as f:
            f.write(text)



"""Test for --gaia-output."""

import os.path

from e3.testsuite import Testsuite as Suite
from e3.testsuite.driver import BasicTestDriver as BasicDriver
from e3.testsuite.result import FailureReason, Log, TestStatus as Status


class TestGAIA:
    class MyDriver(BasicDriver):
        """Dummy driver, push a result built from the test environment."""

        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            status = Status[self.test_env["status"]]
            message = self.test_env.get("message", None)
            self.result.set_status(status, message)

            self.result.failure_reasons = {
                FailureReason[r] for r in self.test_env.get("reasons", [])
            }

            def to_log(key):
                content = self.test_env.get(key)
                return None if content is None else Log(content)

            self.result.log = to_log("log")
            self.result.expected = to_log("expected")
            self.result.out = to_log("out")
            self.result.diff = to_log("diff")

            self.result.time = self.test_env.get("time")
            self.result.info = self.test_env.get("info")

            self.push_result()

    def test(self):
        class Mysuite(Suite):
            test_driver_map = {"mydriver": self.MyDriver}
            default_driver = "mydriver"

        # Run the testsuite, expecting no failure
        suite = Mysuite(os.path.dirname(__file__))
        assert suite.testsuite_main(["--gaia-output"]) == 1

        # Check the content of the GAIA output: first, the summary
        with open(os.path.join("out", "new", "results")) as f:
            lines = sorted(f.read().splitlines())
            assert lines == [
                "fail-binary-out:FAIL:",
                "fail-crash:CRASH:",
                "fail-expected-out-diff:FAIL:",
                "fail-info:FAIL:",
                "fail-memcheck-diff:PROBLEM:",
                "fail-out:FAIL:",
                "fail-simple:FAIL:",
                "fail-timeout:TIMEOUT:",
                "not_applicable:NOT-APPLICABLE:",
                "pass-log:OK:",
                "pass-time:OK:",
                "pass:OK:",
                "skip:DEAD:This test was skipped",
                "xfail-crash:XFAIL:",
                "xfail-out:XFAIL:",
            ]

        # Check the presence and content of log files
        def check_logs(
            test_name,
            *,
            log=None,
            expected=None,
            out=None,
            diff=None,
            time=None,
            info=None
        ):
            for extension, content in [
                (".log", log),
                (".expected", expected),
                (".out", out),
                (".diff", diff),
                (".time", time),
                (".info", info),
            ]:
                filename = os.path.join("out", "new", test_name + extension)
                mode = "r" if isinstance(content, str) else "rb"
                if content is None:
                    assert not os.path.exists(filename)
                else:
                    with open(filename, mode) as f:
                        assert content == f.read()

        check_logs("fail-binary-out", out=b"h\xe9llo")
        check_logs("fail-crash")
        check_logs(
            "fail-expected-out-diff",
            expected="expected\noutput",
            out="actual\noutput",
            diff="-expected\n+actual\n output",
        )
        check_logs("fail-info", info="key1:value1\nkey2:value2")
        check_logs("fail-memcheck-diff")
        check_logs("fail-out", out="output of failed testcase\n")
        check_logs("fail-simple")
        check_logs("fail-timeout")
        check_logs("not_applicable")
        check_logs("pass-log", log="Log for a successful test.\n")
        check_logs("pass-time", time="1.230000000")
        check_logs("pass")
        check_logs("skip")
        check_logs("xfail-crash")
        check_logs("xfail-out", out="Output of failing test in expected way")



"""Helpers to generate testsuite reports using the XUnit XML format."""

from __future__ import annotations

import argparse
import itertools
import os
import re
from typing import Iterable, Optional
import unicodedata
import xml.etree.ElementTree as etree

from e3.fs import mkdir
from e3.testsuite.report.gaia import dump_gaia_report
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import TestResult, TestStatus
import e3.yaml


def add_time_attribute(elt: etree.Element, duration: Optional[float]) -> None:
    """Optionally add a "time" attribute.

    If ``duration`` is a float, add the corresponding "time" attribute to
    ``elt``.
    """
    if duration is not None:
        elt.set("time", "{:0.3f}".format(duration))


def escape_text(text: str) -> str:
    """Escape non-printable characters from a string.

    XML documents cannot contain null or control characters (except newlines).
    """
    result = list(text)
    for i, c in enumerate(result):
        # Replace non-printable characters with their Python escape sequence,
        # but strip quotes.
        if c < " " and c != "\n":
            result[i] = ascii(c)[1:-1]
    return "".join(result)


def dump_xunit_report(name: str, index: ReportIndex, filename: str) -> None:
    """
    Dump a testsuite report to `filename` in the standard XUnit XML format.

    :param name: Name for the teststuite report.
    :param index: Report index for the testsuite results to report.
    :param filename: Name of the text file to write.
    :param duration: Optional number of seconds for the total duration of the
        testsuite run.
    """
    testsuites = etree.Element("testsuites", name=name)
    testsuite = etree.Element("testsuite", name=name)
    testsuites.append(testsuite)

    add_time_attribute(testsuites, index.duration)
    add_time_attribute(testsuite, index.duration)

    # Counters for each category of test in XUnit. We map TestStatus to
    # these.
    counters = {"tests": 0, "errors": 0, "failures": 0, "skipped": 0}
    status_to_counter = {
        TestStatus.PASS: None,
        TestStatus.FAIL: "failures",
        TestStatus.XFAIL: "skipped",
        TestStatus.XPASS: None,
        TestStatus.VERIFY: None,
        TestStatus.SKIP: "skipped",
        TestStatus.NOT_APPLICABLE: "skipped",
        TestStatus.ERROR: "errors",
    }

    # Markup to create inside <testcase> elements for each category of test
    # in XUnit.
    counter_to_markup = {
        "failures": "failure",
        "skipped": "skipped",
        "errors": "error",
    }

    # Now create a <testcase> element for each test
    for test_name, entry in sorted(index.entries.items()):
        result = entry.load()

        # We use the testsuite name for 'classname', as we don't really have
        # something more useful to display anyway.
        testcase = etree.Element("testcase", name=test_name, classname=name)
        testsuite.append(testcase)

        # Get the XUnit-equivalent status for this test and update the
        # corresponding counters.
        counter_key = status_to_counter[result.status]
        if counter_key:
            counters[counter_key] += 1
        counters["tests"] += 1

        # If applicable, create an element to describe the test status. In
        # any case, if we have logs, include them in the report to ease
        # post-mortem debugging. They are included in a standalone
        # "system-out" element in case the test succeeded, or directly in
        # the status element if the test failed.
        markup = counter_key and counter_to_markup.get(counter_key, None)
        if markup:
            status_elt = etree.Element(markup)
            testcase.append(status_elt)
            if counter_key in ("skipped", "errors", "failures") and result.msg:
                status_elt.set("message", escape_text(result.msg))

            if counter_key in ("errors", "failures"):
                status_elt.set("type", "error")

            assert isinstance(result.log, str)
            status_elt.text = escape_text(result.log)

        elif result.log:
            system_out = etree.Element("system-out")
            system_out.text = escape_text(str(result.log))
            testcase.append(system_out)

        add_time_attribute(testcase, result.time)

    # Include counters in <testsuite> and <testsuites> elements
    for key, count in sorted(counters.items()):
        testsuite.set(key, str(count))
        testsuites.set(key, str(count))

    # The report is ready: write it to the requested file
    tree = etree.ElementTree(testsuites)
    tree.write(filename, encoding="utf-8", xml_declaration=True)


class XUnitImporter:
    """Helper class to import results in a xUnit report into a report index."""

    def __init__(
        self,
        index: ReportIndex,
        xfails: dict[str, str] | None = None,
    ):
        """Create a XUnitImporter instance.

        :param index: Report index into which to import xUnit results.
        :param xfails: For each test result that is expected to fail, this dict
            must contain a "test name" to "expected failure message"
            (potentially an empty string) association.
        """
        self.index = index
        self.xfails = xfails or {}

        self.dangling_xfails: set[str] = set()
        """
        Set of tests for which a failure is expected, but that are not present
        in the testsuite report. Computed in the "run" method.
        """

    def warn_dangling_xfails(self) -> None:
        """
        Print warnings for dangling entries in the "xfails" dict.

        This prints warnings on the standard output to mention all tests for
        which a failure is expected, but that are not present in the testsuite
        report.
        """
        if not self.dangling_xfails:
            return
        print(
            "warning: the following tests are expected to fail but are not"
            " present in the testsuite results:"
        )
        for test_name in sorted(self.dangling_xfails):
            reason = self.xfails.get(test_name)
            if reason:
                print(f"  {test_name} ({reason})")
            else:
                print(f"  {test_name}")

    def run(self, filename: str) -> None:
        """Read a xUnit report and import its results in the report index.

        :param filename: Filename for the XML file that contains the xUnit
            report.
        """
        doc = etree.parse(filename)
        testsuites = doc.getroot()
        assert testsuites.tag == "testsuites"

        for testsuite in testsuites:
            assert testsuite.tag == "testsuite"
            testsuite_name = testsuite.attrib["name"]

            for testcase in testsuite:
                # Skip "properties", "system-out" and "system-err" elements so
                # that we process only "testcase" ones.
                if testcase.tag in {"properties", "system-out", "system-err"}:
                    continue
                assert testcase.tag == "testcase"

                testcase_name = testcase.attrib["name"]
                classname = testcase.attrib.get("classname")
                time_str = testcase.attrib.get("time")

                result = TestResult(
                    self.get_test_name(
                        testsuite_name, testcase_name, classname
                    )
                )
                result.time = float(time_str) if time_str else None
                status = TestStatus.PASS
                message: str | None = None

                # Expect at most one "error", "failure" or "skipped" element.
                # Presence of one such element or the absence of elements
                # enables us to determine the test status.
                count = len(testcase)
                assert count <= 1
                if count == 1:
                    status_elt = testcase[0]
                    tag = status_elt.tag
                    if tag == "error":
                        status = TestStatus.ERROR
                    elif tag == "failure":
                        status = TestStatus.FAIL
                    elif tag == "skipped":
                        # xUnit reports created by py.test may contain a "type"
                        # attribute that disambiguates between skip and xfail
                        # results.
                        kind = status_elt.attrib.get("type")
                        status = (
                            TestStatus.XFAIL
                            if kind == "pytest.xfail"
                            else TestStatus.SKIP
                        )
                    else:  # no cover
                        raise AssertionError(f"invalid status tag: {tag}")
                    if isinstance(status_elt.text, str):
                        result.log += status_elt.text

                    message = status_elt.attrib.get("message")

                    # Some XUnit producers are know to put full logs in the
                    # "message" attribute, which produces unexpected results in
                    # e3-testsuite report viewers. Keep the first line only,
                    # and cap its length if needed.
                    capped = False
                    if message is not None:
                        if "\n" in message:
                            message = message.split("\n", 1)[0]
                            capped = True
                        if len(message) > 200:
                            message = message[:200]
                            capped = True
                        if capped:
                            message = message.strip() + " [...]"

                else:
                    message = None

                # Now that the "unrefined" status for this result is known,
                # apply XFAIL, if needed.
                xfail_message = self.xfails.get(result.test_name)
                if xfail_message is not None:
                    # Depending on whether we have an XFAIL message and/or a
                    # test result message, create a single message for the
                    # result to store in the report.
                    new_message = (
                        (
                            f"{message} ({xfail_message})"
                            if message
                            else xfail_message
                        )
                        if xfail_message
                        else message
                    )

                    # xUnit tests often use the ERROR status for issues that
                    # are not testsuite bugs (i.e. for what we call "failures"
                    # in e3-testsuite), so be pragmatic and allow them to be
                    # covered by XFAIL.
                    if status == TestStatus.PASS:
                        status = TestStatus.XPASS
                        message = new_message
                    elif status in (TestStatus.FAIL, TestStatus.ERROR):
                        status = TestStatus.XFAIL
                        message = new_message

                result.set_status(status, message)
                self.index.save_and_add_result(result)

        # Now that all tests are known, compute the set of dangling XFAILs
        self.dangling_xfails = set(self.xfails) - set(self.index.entries)

    SLUG_RE = re.compile("[a-zA-Z0-9_.]+")

    def slugify(self, name: str) -> str:
        """Normalize a string so that it is an acceptable test name component.

        :param name: Component (substring) for a name to turn into a test name
            that is acceptable for e3-testsuite.
        """
        # Normalize the string, decomposing some codepoints into ASCII
        # characters + modifiers
        name = unicodedata.normalize("NFKD", name)

        # Preserve only codepoints in [a-zA-Z0-9_.] and replace/collapse the
        # rest with hyphens.
        return "-".join(chunk for chunk in self.SLUG_RE.findall(name))

    def get_unique_test_name(self, test_name: str) -> str:
        """Return a test name that is guaranteed to be unique.

        :param test_name: Candidate test name. If the report index already has
            a test result with the same test name, this method generates
            another one based on it.
        """
        result = test_name
        counter = itertools.count(1)
        while result in self.index.entries:
            result = f"{test_name}.{next(counter)}"
        return result

    def get_test_name(
        self,
        testsuite_name: str,
        testcase_name: str,
        classname: Optional[str] = None,
    ) -> str:
        """Combine xUnit testsuite/testcase names into a unique test name.

        :param testsuite_name: Name associated with a xUnit <testsuite>
            element.
        :param testcase_name: Name associated with a xUnit <testcase> element.
        :param classname: If applicable, name of the class that owns this
            testcase.
        """
        return self.get_unique_test_name(
            ".".join(
                self.slugify(name)
                for name in [testsuite_name, classname, testcase_name]
                if name
            )
        )


class XUnitImporterApp:
    """
    Helper class to implement a xUnit report import script.

    This class provide the basic behavior, which subclasses can override if
    needed.
    """

    def __init__(self) -> None:
        self.parser = argparse.ArgumentParser(
            description="Convert a xUnit testsuite report to e3-testsuite's"
            " format."
        )
        self.add_basic_options(self.parser)
        self.add_options(self.parser)

        self.args: argparse.Namespace
        self.index: ReportIndex
        self.xfails: dict[str, str]
        self.importer: XUnitImporter

    def add_basic_options(self, parser: argparse.ArgumentParser) -> None:
        """
        Add basic command line arguments.

        Subclasses must override this to replace basic command line arguments.
        Note that most methods/properties in this class assume that these
        arguments are present: not adding them will likely require to override
        all methods/properties.
        """
        parser.add_argument(
            "--output",
            "-o",
            help="Output directory for the converted report. By default, use"
            " the current working directory.",
        )
        parser.add_argument(
            "--gaia-output",
            action="store_true",
            help="Output a GAIA-compatible testsuite report next to the YAML"
            " report.",
        )
        parser.add_argument(
            "--xfails",
            help="YAML file that describes expected failures. If provided, it"
            " must contain a mapping from test name to expected failure"
            " messages.",
        )
        parser.add_argument(
            "xml-report",
            nargs="+",
            help="xUnit XML reports to convert. If a directory is passed,"
            " recursively look for all the files matching *.xml that it"
            " contains.",
        )

    def add_options(self, parser: argparse.ArgumentParser) -> None:
        """
        Add extra command line arguments.

        Subclasses must override this to add command line arguments in addition
        to the basic ones.
        """
        pass

    @property
    def output_dir(self) -> str:
        """Return the report output directory."""
        return self.args.output or "."

    def create_output_report_index(self) -> ReportIndex:
        """Create the index for the report this app must write."""
        result = ReportIndex(self.output_dir)
        mkdir(result.results_dir)
        return result

    def get_xfails(self) -> dict[str, str]:
        """Return the "xfails" XUnitImporter constructor argument."""
        return (
            read_xfails_from_yaml(self.args.xfails) if self.args.xfails else {}
        )

    def create_importer(self) -> XUnitImporter:
        """Create the XUnitImporter instance for this app."""
        return XUnitImporter(self.index, self.xfails)

    def iter_xunit_files(self) -> Iterable[str]:
        """Iterate through all the xUnit report files to import."""
        for path in getattr(self.args, "xml-report"):
            if os.path.isdir(path):
                for root, _, filenames in os.walk(path):
                    for f in filenames:
                        yield os.path.join(root, f)
            else:
                yield path

    @property
    def gaia_report_requested(self) -> bool:
        """Return whether a GAIA report was requested."""
        return self.args.gaia_output

    def tear_down(self) -> None:
        """
        Clean up the importer.

        Subclasses must override this for custom behavior before the app ends.
        """
        pass

    def run(self, argv: list[str] | None = None) -> None:
        # Initialize the importer
        self.args = self.parser.parse_args(argv)
        self.index = self.create_output_report_index()
        self.xfails = self.get_xfails()
        self.importer = self.create_importer()

        # Process xUnit report files
        for filename in self.iter_xunit_files():
            self.importer.run(filename)
        self.importer.warn_dangling_xfails()

        # Write reports
        self.index.write()
        if self.gaia_report_requested:
            dump_gaia_report(self.index, self.index.results_dir)

        # Clean up the importer
        self.tear_down()


def read_xfails_from_yaml(filename: str) -> dict[str, str]:
    """
    Read a XFAILs dict from a YAML file.

    See the "xfails" parameter for XUnitImporter's constructor for the expected
    YAML structure.
    """
    return e3.yaml.load_ordered(filename)


def convert_main(argv: list[str] | None = None) -> None:
    XUnitImporterApp().run(argv)



"""Data structures for testcase execution results."""

from __future__ import annotations

import binascii
from dataclasses import dataclass
from enum import Enum, auto
import logging
import os.path
import tempfile
from typing import Any, AnyStr, Dict, Iterator, Generic, Optional, Set, cast

import yaml

from e3.testsuite.utils import ColorConfig


class TestStatus(Enum):
    """Testcase execution status."""

    # The test has run to completion and has succeeded
    PASS = auto()

    # The test has run enough for the testsuite to consider that it failed
    FAIL = auto()

    # The test has run enough for the testsuite to consider that it failed, and
    # that this failure was expected.
    XFAIL = auto()

    # The test has run to completion and has succeeded whereas a failure was
    # expected.
    XPASS = auto()

    # The test has run to completion, but it could not self-verify the test
    # objective (i.e. determine whether it succeeded). This test requires an
    # additional verification action by a human or some external oracle.
    VERIFY = auto()

    # The test was not executed (it has been skipped). This is appropriate when
    # the test does not make sense in the current configuration (for instance
    # it must run on Windows, and the current OS is GNU/Linux).
    #
    # This is equivalent to DejaGnu's UNSUPPORTED, or UNTESTED test outputs.
    SKIP = auto()

    # The test has run and managed to automatically determine it can't work on
    # a given configuration (for instance, a test scenario requires two
    # distinct interrupt priorities, but only one is supported on the current
    # target).
    #
    # The different with SKIP is that here, the test has started when it
    # determined that it would not work. The definition of when a test actually
    # starts is left to the test driver.
    NOT_APPLICABLE = auto()

    # The test could not run to completion because it is misformatted or due to
    # an unknown error. This is very different from FAIL, because here the
    # problem comes more likely from the testcase or the test framework rather
    # than the tested software.
    #
    # This is equivalent to DejaGnu's UNRESOLVED test output.
    ERROR = auto()

    def color(self, colors: ColorConfig) -> str:
        """Return the ANSI color code for this test status.

        This returns an empty string if colors are disabled.
        """
        Fore = colors.Fore
        Style = colors.Style
        return {
            "PASS": Fore.GREEN,
            "FAIL": Fore.RED,
            "XFAIL": Fore.CYAN,
            "XPASS": Fore.YELLOW,
            "VERIFY": Fore.YELLOW,
            "SKIP": Style.DIM,
            "NOT_APPLICABLE": Style.DIM,
            "ERROR": Fore.RED + Style.BRIGHT,
        }[self.name]


class FailureReason(Enum):
    """Reason for a test failure."""

    # A process crash was detected. What is a "crash" is not clearly specified:
    # it could be for instance that a "GCC internal compiler error" message is
    # present in the test output.
    CRASH = auto()

    # A process was stopped because it timed out
    TIMEOUT = auto()

    # The tested software triggered an invalid memory access pattern. For
    # instance, Valgrind found a conditional jump that depends on uninitialized
    # data.
    MEMCHECK = auto()

    # Output is not as expected
    DIFF = auto()


class Log(yaml.YAMLObject, Generic[AnyStr]):
    """Object to hold long text or binary logs.

    We ensure that when dump to yaml the result will be human readable.
    """

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!e3.testsuite.result.Log"

    def __init__(self, content: AnyStr) -> None:
        """Initialize log instance.

        :param content: Initial message to log.
        """
        assert isinstance(content, (str, bytes))
        self.log: AnyStr = content

    @property
    def is_binary(self) -> bool:
        """Return whether this log contains binary data."""
        return isinstance(self.log, bytes)

    @property
    def is_text(self) -> bool:
        """Return whether this log contains text data."""
        return not self.is_binary

    def __iadd__(self, content: AnyStr) -> Log[AnyStr]:
        """Add additional content to the log.

        :param content: a message to log
        """
        self.log += content
        return self

    def __str__(self) -> str:
        return (
            cast(str, self.log)
            if self.is_text
            else binary_repr(cast(bytes, self.log))
        )


def binary_repr(binary: bytes) -> str:
    r"""Return a human readable representation for the given bytes string.

    This just decodes ASCII printable bytes and newlines to the corresponding
    strings and represents other bytes with the "\xXX" escapes.

    :param binary: Bytes string to represent.
    """

    def escape(b: int) -> str:
        if b == ord("\\"):
            return "\\\\"
        elif b == ord("\n") or (b >= ord(" ") and b <= ord("~")):
            return chr(b)
        else:
            return "\\x{:>02x}".format(b)

    return "\n".join(
        "".join(escape(b) for b in line) for line in binary.split(b"\n")
    )


def truncated(output: str, line_count: int) -> str:
    """Truncate an output not to exceed twice the given number of lines.

    If ``output`` has more than ``2 * line_count`` lines, only keep the first
    ``N`` and last ``N`` lines of it. Return it unchanged otherwise, or if
    ``line_count`` is 0.

    :param output: Output to (maybe) truncate.
    :param line_count: Half the maximum number of lines to keep.
    """
    # Given that we insert Unix-style line breaks here, make .splitlines()
    # strip line terminators. If ``output`` contains Windows-style ones, this
    # function will convert them to Unix-style ones. This transformation
    # mangles the output a bit, but this is necessary to keep things tractable
    # here.
    lines = output.splitlines()
    max_lines = 2 * line_count
    if line_count and len(lines) > max_lines:
        lines = (
            lines[:line_count]
            + ["\n... {} lines skipped...\n".format(len(lines) - max_lines)]
            + lines[-line_count:]
        )
    return "\n".join(lines)


class TestResult(yaml.YAMLObject):
    """Represent a result for a given test."""

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!e3.testsuite.result.TestResult"

    def __init__(
        self,
        name: str,
        env: Optional[dict] = None,
        status: Optional[TestStatus] = None,
        msg: str = "",
    ):
        """Initialize a test result.

        :param name: Name of the test that this result describes.
        :param env: Test environment. Usually a dict that contains relevant
            test information (output, ...). The object should be serializable
            to YAML format.
        :param status: Test status. If None status is set to ERROR.
        :param msg: Short message associated with the test result.
        """
        self.test_name = name
        self.env = env

        # Use the set_status method to change these once initialization is done
        self.status = TestStatus.ERROR if status is None else status
        self.msg: Optional[str] = msg

        # Free-form text, for test execution post-mortem debugging purposes.
        # Test drivers are invited to write content that will be useful if
        # things go wrong during the test execution: test failure, test driver
        # bug, and so on. This is what gets printed on the standard output when
        # the test fails and the "--show-error-output" testsuite switch is
        # present.
        self.log = Log("")

        # List of free-form information (but still serializable to YAML)
        # describing the subprocesses that the test driver spawned while runnig
        # the testcase.
        #
        # Note that both e3.testsuite.process.check_call and
        # e3.testsuite.driver.classic.ClassicTestDriver.shell fill this
        # automatically with command-line arguments, exit code, etc.
        self.processes: list = []

        # When the test failed, optional set of reasons for the failure. This
        # information is used only in advanced viewers, which may highlight
        # specifically some failure reasons. For instance, highlight crashes,
        # that may be more important to investigate than mere unexpected
        # outputs.
        self.failure_reasons: Set[FailureReason] = set()

        # Drivers that compare expected and actual output to validate a
        # testcase should initialize these with Log instances to hold the
        # expected test output (self.expected) and the actual test output
        # (self.out). It is assumed that the test fails when there is at least
        # one difference between both.
        #
        # Note that several drivers refine expected/actual outputs before
        # running the comparison (see for instance the
        # e3.testsuite.driver.diff.OutputRefiner mechanism). These logs are
        # supposed to contain the outputs actually passed to the diff
        # computation function, i.e. *after* refining, so that whatever attemps
        # to re-compute the diff (report production, for instance) get the same
        # result.
        #
        # If, for some reason, it is not possible to store expected and actual
        # outputs, self.diff can be assigned a Log instance holding the diff
        # itself. For instance, the output of the `diff -u` command.
        self.expected: Optional[Log] = None
        self.out: Optional[Log] = None
        self.diff: Optional[Log] = None

        # Optional decimal number of seconds (float). Test drivers can use this
        # field to track performance, most likely the time it took to run the
        # test. Advanced results viewer can then plot the evolution of time
        # over software evolution.
        self.time: Optional[float] = None

        # Key/value string mapping, for unspecified use. The only restriction
        # is that no string can contain a newline character.
        self.info: Dict[str, str] = {}

    def set_status(self, status: TestStatus, msg: Optional[str] = "") -> None:
        """Update the test status.

        :param status: New status. Note that only test results with status set
            to ERROR can be changed.
        :param msg: Optional short message to describe the result.  Note that
            multiline strings are turned into single-line strings.
        """
        if self.status != TestStatus.ERROR:
            logging.error("cannot set test %s status twice", self.test_name)
            return
        self.status = status

        self.msg = (
            " ".join(line.strip() for line in msg.splitlines() if line)
            if msg
            else None
        )

    def __str__(self) -> str:
        return "%-24s %-12s %s" % (self.test_name, self.status, self.msg)

    def save(self, results_dir: str) -> str:
        """Write this test results as a YAML file.

        :param results_dir: Name of the directory in which to write the test
            result. When writing a testsuite report, this corresponds to the
            report's ``results_dir`` (see
            ``e3.testsuite.report.index.ReportIndex``).
        :return: The base filename of the file written. It is generated from
            the testname.
        """
        with tempfile.NamedTemporaryFile(
            mode="w",
            prefix=self.test_name + "-",
            suffix=".yaml",
            dir=results_dir,
            delete=False,
        ) as f:
            yaml.dump(self, f)
            return os.path.basename(f.name)

    @property
    def summary(self) -> TestResultSummary:
        return TestResultSummary(
            self.test_name,
            self.status,
            self.msg,
            self.failure_reasons,
            self.time,
            self.info,
        )


# Enforce representation of Log objects when dumped to yaml
def _log_representer(dumper: Any, data: Log) -> Any:
    return (
        dumper.represent_scalar("tag:yaml.org,2002:str", data.log, style="|")
        if data.is_text
        else dumper.represent_scalar(
            "tag:yaml.org,2002:binary",
            binascii.b2a_base64(data.log).decode("ascii"),
            style="|",
        )
    )


yaml.add_representer(Log, _log_representer)


_test_status_tag = "!e3.testsuite.result.TestStatus"
_failure_reason_tag = "!e3.testsuite.result.FailureReason"


# We cannot use yaml.YAMLObject metaclass magic for TestStatus as it derives
# from Enum, which already has a metaclass. So use an alternative YAML API to
# make it serializable. Likewise for FailureReason.
def _test_status_constructor(self: Any, node: Any) -> Iterator[Any]:
    # Get the numeric value corresponding to the test status, then build a
    # TestStatus instance from it.
    num = int(node.value[0])
    status = TestStatus(num)
    yield status


def _test_status_representer(dumper: Any, data: TestStatus) -> Any:
    return dumper.represent_scalar(
        _test_status_tag, str(data.value), style="|"
    )


def _failure_reason_constructor(self: Any, node: Any) -> Iterator[Any]:
    # Get the numeric value corresponding to the test status, then build a
    # TestStatus instance from it.
    num = int(node.value[0])
    status = FailureReason(num)
    yield status


def _failure_reason_representer(dumper: Any, data: FailureReason) -> Any:
    return dumper.represent_scalar(
        _failure_reason_tag, str(data.value), style="|"
    )


yaml.SafeLoader.add_constructor(_test_status_tag, _test_status_constructor)
yaml.add_representer(TestStatus, _test_status_representer)
yaml.SafeLoader.add_constructor(
    _failure_reason_tag, _failure_reason_constructor
)
yaml.add_representer(FailureReason, _failure_reason_representer)


@dataclass(frozen=True)
class TestResultSummary:
    """Basic data about a test result.

    This class holds a subset of the information available in ``TestResult``
    instances. This subset is meant to be small enough to be included in result
    indexes while enabling efficient common processings such as "show logs for
    tests that failed": no need to load all the test results for tests that
    succeeded.
    """

    test_name: str
    status: TestStatus
    msg: Optional[str]
    failure_reasons: Set[FailureReason]
    time: Optional[float]
    info: Dict[str, str]



"""Tests for the e3.testsuite.report.rewriting module."""

import os

from e3.testsuite.report.gaia import dump_gaia_report
from e3.testsuite.report.rewriting import (
    BaseBaselineRewriter,
    RewritingError,
    RewritingSummary,
)
from e3.testsuite.result import FailureReason, TestStatus as Status
from e3.testsuite.utils import ColorConfig

from .utils import create_report, create_result


def create_initial_baselines(tests):
    return {
        test_name: f"Initial baseline for {test_name}" for test_name in tests
    }


# The baselines to update
initial_baselines = create_initial_baselines(
    [
        "t-pass",
        "t-xfail",
        "t-error",
        "t-fail-no-reason",
        "t-fail-diff",
        "t-fail-diff-crash",
        "t-bytes",
        "t-utf-8",
        "t-out-empty",
        "t-out-none",
    ]
)
initial_baselines["t-no-baseline"] = ""

# The test results used to update baselines
test_results = {
    create_result("t-pass", Status.PASS, out="INVALID"),
    create_result("t-xfail", Status.XFAIL, out="INVALID"),
    create_result("t-error", Status.ERROR, out="INVALID"),
    create_result("t-fail-no-reason", Status.FAIL, out="INVALID"),
    create_result(
        "t-fail-diff",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out="New baseline",
    ),
    create_result(
        "t-fail-diff-crash",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF, FailureReason.CRASH},
        out="INVALID",
    ),
    create_result(
        "t-bytes",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out=b"\xff\x00\x80",
    ),
    create_result(
        "t-utf-8",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out="\xe9",
    ),
    create_result(
        "t-latin-1",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out="\xe9",
        encoding="latin-1",
    ),
    create_result(
        "t-out-empty",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out=b"",
    ),
    create_result(
        "t-out-none",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
    ),
    create_result(
        "t-no-baseline",
        Status.FAIL,
        failure_reasons={FailureReason.DIFF},
        out="New baseline",
    ),
}

# The baselines expected after the update
updated_baselines = {
    "t-pass": b"Initial baseline for t-pass",
    "t-xfail": b"Initial baseline for t-xfail",
    "t-error": b"Initial baseline for t-error",
    "t-fail-no-reason": b"Initial baseline for t-fail-no-reason",
    "t-fail-diff": b"New baseline",
    "t-fail-diff-crash": b"Initial baseline for t-fail-diff-crash",
    "t-bytes": b"\xff\x00\x80",
    "t-utf-8": b"\xc3\xa9",
    "t-latin-1": b"\xe9",
    "t-out-empty": b"",
    "t-out-none": b"Initial baseline for t-out-none",
    "t-no-baseline": b"New baseline",
}

expected_summary = RewritingSummary(
    errors={
        "t-error",
        "t-fail-no-reason",
        "t-fail-diff-crash",
        "t-out-none",
    },
    updated_baselines={"t-fail-diff", "t-bytes", "t-utf-8"},
    new_baselines={"t-latin-1", "t-no-baseline"},
    deleted_baselines={"t-out-empty"},
)

# The GAIA format is less precise than the default one: its processing yields
# slightly different results.
gaia_updated_baselines = dict(updated_baselines)
gaia_expected_summary = RewritingSummary(
    set(expected_summary.errors),
    set(expected_summary.updated_baselines),
    set(expected_summary.new_baselines),
    set(expected_summary.deleted_baselines),
)

# The GAIA format cannot distinguish the "no output" case from the "empty
# output" one.
gaia_updated_baselines["t-out-none"] = b""

# The encoding information is also lost in GAIA reports, so we have to assume
# that everything is UTF-8.
gaia_updated_baselines["t-latin-1"] = b"\xc3\xa9"

# Some errors cannot be distinguished properly from actual diffs
for t in ("t-fail-no-reason", "t-fail-diff-crash", "t-out-none"):
    gaia_expected_summary.errors.remove(t)
gaia_expected_summary.deleted_baselines.add("t-out-none")


def baseline_filename(tmp_path, test_name):
    return str(tmp_path / "baselines" / f"{test_name}.out")


def write_baselines(tmp_path, baselines):
    """Write initial baselines under "tmp_path/baselines"."""
    (tmp_path / "baselines").mkdir()
    for test_name, baseline in baselines.items():
        if baseline:
            with open(baseline_filename(tmp_path, test_name), "w") as f:
                f.write(baseline)


def check_baselines(tmp_path, baselines):
    """Check actual baselines against expectations."""
    for test_name, baseline in baselines.items():
        filename = baseline_filename(tmp_path, test_name)
        if baseline:
            with open(filename, "rb") as f:
                assert f.read() == baseline
        else:
            assert not os.path.exists(filename)


class BR(BaseBaselineRewriter):
    """Simple implementation for a baseline rewriter."""

    def __init__(self, tmp_path):
        self.tmp_path = tmp_path
        super().__init__(ColorConfig(colors_enabled=False))

    def baseline_filename(self, test_name):
        return baseline_filename(self.tmp_path, test_name)


def do_setup(tmp_path, initial_baselines, test_results, br_cls=BR):
    """Set up baselines and a testsuite report."""
    write_baselines(tmp_path, initial_baselines)
    br = br_cls(tmp_path)
    results_dir = tmp_path / "results"
    results_dir.mkdir()
    return br, create_report(test_results, str(results_dir))


def test_report(tmp_path):
    """Test baseline updates from a "native" e3-testsuite report index."""
    br, report = do_setup(tmp_path, initial_baselines, test_results)
    assert br.rewrite(report.results_dir) == expected_summary
    check_baselines(tmp_path, updated_baselines)


def test_deleted_baseline(tmp_path):
    """Test baseline deletion when the baseline is already missing."""
    br, report = do_setup(tmp_path, initial_baselines, test_results)
    # Remove the baseline
    os.remove(os.path.join(tmp_path, "baselines", "t-out-empty.out"))
    deleted_expected_summary = RewritingSummary(
        set(expected_summary.errors),
        set(expected_summary.updated_baselines),
        set(expected_summary.new_baselines),
        set(),
    )
    assert br.rewrite(report.results_dir) == deleted_expected_summary
    check_baselines(tmp_path, updated_baselines)


def test_gaia(tmp_path):
    """Test baseline updates from a GAIA report."""
    br, report = do_setup(tmp_path, initial_baselines, test_results)
    gaia_dir = tmp_path / "gaia"
    gaia_dir.mkdir()
    dump_gaia_report(report, gaia_dir)
    assert br.rewrite(str(gaia_dir)) == gaia_expected_summary
    check_baselines(tmp_path, gaia_updated_baselines)


def test_no_report(tmp_path):
    """Check that no report found is properly reported."""
    br = BR(tmp_path)
    try:
        br.rewrite(str(tmp_path))
    except RewritingError as exc:
        br.print_error(str(exc))
    else:
        raise AssertionError()


def test_gaia_short_status(tmp_path):
    """Check that GAIA's special single-letter ".result" files are handled."""
    initial_baselines = create_initial_baselines(
        ["t-pass", "t-fail", "t-problem", "t-diff"]
    )
    test_results = {
        create_result("t-pass", Status.PASS, out="INVALID"),
        create_result("t-fail", Status.FAIL, out="INVALID"),
        create_result("t-problem", Status.ERROR, out="INVALID"),
        create_result(
            "t-diff",
            Status.FAIL,
            failure_reasons={FailureReason.DIFF},
            out="New baseline",
        ),
    }
    br, report = do_setup(tmp_path, initial_baselines, test_results)

    updated_baselines = {
        "t-pass": b"Initial baseline for t-pass",
        "t-fail": b"Initial baseline for t-fail",
        "t-problem": b"Initial baseline for t-problem",
        "t-diff": b"New baseline",
    }

    # Create the GAIA report and patch the "*.result" files
    gaia_dir = tmp_path / "gaia"
    gaia_dir.mkdir()
    dump_gaia_report(report, gaia_dir)
    for test_name, letter in [
        ("t-pass", "O"),
        ("t-fail", "I"),
        ("t-problem", "C"),
        ("t-diff", "D"),
    ]:
        with (tmp_path / "gaia" / f"{test_name}.result").open("w") as f:
            f.write(letter)

    assert br.rewrite(str(gaia_dir)) == RewritingSummary(
        errors={"t-problem"},
        updated_baselines={"t-diff"},
        new_baselines=set(),
        deleted_baselines=set(),
    )
    check_baselines(tmp_path, updated_baselines)


def test_baseline_postprocessing(tmp_path):
    """Check that baseline postprocessing works as expected."""

    class PPBR(BR):
        def postprocess_baseline(self, baseline):
            return baseline.decode("ascii").lower().encode("ascii")

    initial_baselines = create_initial_baselines(["t"])
    test_results = {
        create_result(
            "t",
            Status.FAIL,
            failure_reasons={FailureReason.DIFF},
            out="New Baseline",
        ),
    }
    updated_baselines = {"t": b"new baseline"}

    br, report = do_setup(tmp_path, initial_baselines, test_results, PPBR)
    assert br.rewrite(report.results_dir) == RewritingSummary(
        errors=set(),
        updated_baselines={"t"},
        new_baselines=set(),
        deleted_baselines=set(),
    )

    check_baselines(tmp_path, updated_baselines)



from __future__ import annotations

import abc
import argparse
from dataclasses import dataclass
import os.path
import traceback
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import e3.collection.dag
import e3.env

from e3.testsuite.report.gaia import (
    GAIAResultFiles,
    dump_result_logs_if_needed,
)
from e3.testsuite.result import TestResult, TestResultSummary
from e3.testsuite.utils import CleanupMode

if TYPE_CHECKING:
    from e3.testsuite.fragment import FragmentCallback


@dataclass(frozen=True)
class ResultQueueItem:
    """Information to integrate a test result in a testsuite report.

    Test drivers create test results. They travel from there to the testsuite
    final report through various queues. This class gathers all the information
    needed for the various stages of this pipeline.
    """

    result: TestResultSummary
    """Summary for this test result."""

    filename: str
    """Name of the file that contains test result data."""

    traceback: List[str]
    """Stack trace for this result's push time.

    This stack trace corresponds to the code that led the TestResult instance
    to be included in the testsuite report.
    """

    gaia_results: Optional[GAIAResultFiles]
    """GAIA files for this result.

    This is None if no GAIA report is requested.
    """


ResultQueue = List[ResultQueueItem]


class TestDriver(object, metaclass=abc.ABCMeta):
    """Testsuite Driver.

    All drivers declared in a testsuite should inherit from this class
    """

    Fore: Any
    Style: Any

    test_env_filename: str

    def __init__(self, env: e3.env.Env, test_env: Dict[str, Any]) -> None:
        """Initialize a TestDriver instance.

        :param env: The testsuite environment. This mirrors the
            ``Testsuite.env`` attribute.
        :param test_env: The testcase environment. By the time it is passed to
            this constructor, a TestFinder subclass has populated it, and the
            testsuite added the following entries:

            * ``test_dir``: The absolute name of the directory that contains
              the testcase.
            * ``test_name``: The name that the testsuite assigned to this
              testcase.
            * ``working_dir``: The absolute name of the temporary directory
              that this test driver is free to create (if needed) in order to
              run the testcase.
        """
        self.env: e3.env.Env = env
        assert isinstance(env.options, argparse.Namespace)
        self.testsuite_options: argparse.Namespace = env.options

        self.test_env: Dict[str, Any] = test_env
        self.test_name: str = test_env["test_name"]

        # Initialize test result
        self.result: TestResult = TestResult(
            name=self.test_name, env=self.test_env
        )

        self.result_queue: ResultQueue = []
        """
        Queue of test results that this driver plans to integrate to the
        testsuite report.
        """

    def push_result(self, result: Optional[TestResult] = None) -> None:
        """Push a result to the testsuite.

        This method should be called to push results to the testsuite report.

        :param result: A TestResult object to push. If None push the current
            test result.
        """
        if result is None:
            result = self.result

        # Write the result as a new file in the output directory and let the
        # testsuite main know about this result.
        output_dir = self.env.output_dir
        assert isinstance(output_dir, str)

        self.result_queue.append(
            ResultQueueItem(
                result.summary,
                result.save(self.env.output_dir),
                traceback.format_stack(),
                dump_result_logs_if_needed(self.env, result, output_dir),
            )
        )

    def add_fragment(
        self,
        dag: e3.collection.dag.DAG,
        name: str,
        fun: Optional[FragmentCallback] = None,
        after: Optional[List[str]] = None,
    ) -> None:
        """Add a test fragment.

        This function is a helper to define test workflows that do not
        introduce dependencies to other tests. For more complex operation use
        directly add_vertex method from the dag. See add_test method.

        :param dag: DAG containing test fragments.
        :param name: Name of the fragment.
        :param fun: Callable that takes two positional arguments: a mapping
            from fragment names to return values for already executed
            fragments, and a slot ID. If None looks for a method inside this
            class called ``name``.
        :param after: List of fragment names that should be executed before
            this one.
        """
        from e3.testsuite.fragment import FragmentData

        callback_by_name = False

        if after is not None:
            after = [self.test_name + "." + k for k in after]

        if fun is None:
            fun = getattr(self, name)
            callback_by_name = True

        fragment = FragmentData(
            uid=f"{self.test_name}.{name}",
            driver=self,
            name=name,
            callback=fun,
            callback_by_name=callback_by_name,
        )

        dag.update_vertex(
            vertex_id=fragment.uid,
            data=fragment,
            predecessors=after,
            enable_checks=False,
        )

    def working_dir(self, *args: str) -> str:
        """Build a filename in the test working directory."""
        return os.path.join(self.test_env["working_dir"], *args)

    def test_dir(self, *args: str) -> str:
        """Build a filename in the testcase directory."""
        return os.path.join(self.test_env["test_dir"], *args)

    @abc.abstractmethod
    def add_test(self, dag: e3.collection.dag.DAG) -> None:
        """Create the test workflow.

        Amend a DAG with the test fragments that should be executed along with
        their dependencies. See BasicTestDriver for an example of workflow.
        """
        raise NotImplementedError

    @property
    def working_dir_cleanup_enabled(self) -> bool:
        """
        Return whether test drivers should cleanup their working directory.

        Unless this returns False, test drivers should delete their working
        directory when the test has completed, so that temporaries for the
        whole testsuite are removed incrementally. This is necessary to avoid
        creating huge temporary directories when executing big testsuites.
        """
        return self.env.cleanup_mode != CleanupMode.NONE


class BasicTestDriver(TestDriver, metaclass=abc.ABCMeta):
    def add_test(self, dag: e3.collection.dag.DAG) -> None:
        """Create a standard test workflow.

        set up -> run -> analyze -> tear_down in which set up and tear_down
        are optional.
        """
        self.add_fragment(dag, "set_up")
        self.add_fragment(dag, "run", after=["set_up"])
        self.add_fragment(dag, "analyze", after=["run"])
        self.add_fragment(dag, "tear_down", after=["analyze"])

    def set_up(self, prev: Dict[str, Any], slot: int) -> None:
        """Execute operations before executing a test."""
        return self.tear_up(prev, slot)

    def tear_up(self, prev: Dict[str, Any], slot: int) -> None:
        """Backwards-compatible name for the "set_up" method."""
        pass

    def tear_down(self, prev: Dict[str, Any], slot: int) -> None:
        """Execute operations once a test is finished."""
        pass

    @abc.abstractmethod
    def run(self, prev: Dict[str, Any], slot: int) -> None:
        """Execute a test."""
        raise NotImplementedError

    @abc.abstractmethod
    def analyze(self, prev: Dict[str, Any], slot: int) -> None:
        """Compute the test result."""
        raise NotImplementedError



import os
import sys


crlf = False


def print_line(line):
    sys.stdout.write(line)
    sys.stdout.write("\r\n" if crlf else "\n")


for arg in sys.argv[1:]:
    if arg == "-e":
        sys.exit(1)
    if arg == "-crlf":
        crlf = True
    elif arg == "-cwd":
        print_line(os.getcwd())
    elif arg.startswith("-read="):
        with open(arg[6:], "r"):
            pass
    elif arg.startswith("-p"):
        print_line(arg[2:])
    elif arg.startswith("-b"):
        sys.stdout.buffer.write(b"h\xe9llo")
    elif arg.startswith("-i"):
        sys.stdout.buffer.write("héllo".encode("iso-8859-1"))
    elif arg == "-stdin":
        print("From stdin:", repr(sys.stdin.read()))
    else:
        print("Unknown switch")
        sys.exit(2)



from __future__ import annotations

"""
Test fragments handling.

Test drivers can split the execution of their test in multiple test fragments.
Each fragment is an atomic task, which can be dispatched to a separate worker.
"""

import abc
import argparse
from dataclasses import dataclass
import pickle
import sys
import tempfile
import traceback
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    TYPE_CHECKING,
    Type,
)

import e3.env
from e3.fs import rm
import e3.job
from e3.os.process import DEVNULL, PIPE, Run, STDOUT
from e3.testsuite.driver import ResultQueue, TestDriver
import e3.testsuite.multiprocess_scheduler
from e3.testsuite.result import TestResult, TestStatus


if TYPE_CHECKING:
    from e3.testsuite.running_status import RunningStatus


class FragmentCallback(Protocol):
    def __call__(self, previous_values: Dict[str, Any], slot: int) -> None: ...


@dataclass
class FragmentData:
    """Data for a job unit in the testsuite.

    Each ``FragmentData`` instance is recorded in the testsuite global DAG to
    control the order of execution of all fragments with the requested level of
    parallelism.

    Note that the job scheduler turns ``FragmentData`` instances into
    ``TestFragment`` ones during the execution (see ``Testsuite.job_factory``
    callback).
    """

    uid: str
    driver: TestDriver
    name: str
    callback: FragmentCallback

    callback_by_name: bool
    """Whether ``callback`` is just the ``name`` method of ``driver``."""

    def matches(self, driver_cls: Type[TestDriver], name: str) -> bool:
        """Return whether this fragment matches the given name/test driver.

        If ``name`` is left to None, just check the driver type.
        """
        return isinstance(self.driver, driver_cls) and (
            name is None or self.name == name
        )

    def clear_driver_data(self) -> None:
        """Remove references to ``TestDriver`` instances and related data.

        Doing this is necessary after each fragment is complete to keep memory
        consumption under control for big testsuites: test driver instances may
        contain a lot of data.
        """
        joker: Any = None
        self.driver = joker
        self.callback = joker


class TestFragment:
    """Base class for testcase scheduling units."""

    uid: str
    """Unique string identifier for this test fragment."""

    driver: TestDriver
    """Test driver that is responsible for this test fragment."""

    running_status: RunningStatus
    """RunningStatus instance to signal when job starts/completes."""

    result_queue: ResultQueue
    """
    List of test results that this fragments plans to integrate to the
    testsuite report.
    """

    @staticmethod
    def static_push_error_result(uid: str, driver: TestDriver) -> None:
        """Generate a test result to log the exception and traceback.

        This helper method is meant to be used when the execution of the test
        fragments aborts because of an uncaught exception. We must report a
        test error, and we provide exception information for post-mortem
        investigation.

        :param uid: UID for the test fragment.
        :param driver: TestDriver for the test fragment.
        """
        # Create a result whose name matches the fragment's uid. If the main
        # testsuite loop's result collection logic detects a name conflict, it
        # will automatically add a suffix to that name.
        result = TestResult(uid, env=driver.test_env, status=TestStatus.ERROR)
        result.log += traceback.format_exc()
        driver.push_result(result)

    def push_error_result(self, exc: Exception) -> None:
        """Shortcut for static_push_error_result on the current fragment."""
        self.static_push_error_result(self.uid, self.driver)

    def must_run(self) -> bool:
        """Return if the fragment must be run.

        It must be skipped when the testsuite has aborted because of too many
        consecutive failures: we push a SKIP result in that case.
        """
        if self.running_status.aborted_too_many_failures:
            self.driver.push_result(
                TestResult(
                    self.uid,
                    env=self.driver.test_env,
                    status=TestStatus.SKIP,
                    msg="Testsuite aborted",
                )
            )
            return False
        else:
            return True

    @abc.abstractmethod
    def clear_driver_data(self) -> None:
        """Remove references to ``TestDriver`` instances and related data.

        Doing this is necessary after each fragment is complete to keep memory
        consumption under control for big testsuites: test driver instances may
        contain a lot of data.
        """
        ...


class ThreadTestFragment(e3.job.Job, TestFragment):
    """Run a test fragment in a thread."""

    def __init__(
        self,
        uid: str,
        driver: TestDriver,
        callback: FragmentCallback,
        previous_values: Dict[str, Any],
        notify_end: Callable[[str], None],
        running_status: RunningStatus,
    ) -> None:
        """Initialize a TestFragment.

        :param uid: UID of the test fragment (should be unique).
        :param driver: A TestDriver instance.
        :param callback: Callable to be executed by the job.
        :param notify_end: Internal parameter. See e3.job.
        :param running_status: See BaseTestFragment.running_status.
        """
        super().__init__(uid, callback, notify_end)
        self.driver = driver
        self.previous_values = previous_values
        self.running_status = running_status
        self.result_queue: ResultQueue = driver.result_queue

    def clear_driver_data(self) -> None:
        joker: Any = None
        self.data = joker
        self.driver = joker

    def run(self) -> None:
        """Run the test fragment."""
        from e3.testsuite import TestAbort

        self.return_value = None
        if self.must_run():
            self.running_status.start(self)
            try:
                self.return_value = self.data(self.previous_values, self.slot)
            except TestAbort:
                pass
            except Exception as e:
                self.push_error_result(e)
                self.return_value = e
            self.running_status.complete(self)


class ProcessTestFragment(
    e3.testsuite.multiprocess_scheduler.Worker, TestFragment
):
    """Run a test fragment in a separate process."""

    @dataclass(frozen=True)
    class Input:
        """Subprocess input data."""

        fragment_uid: str
        driver_cls: Type[TestDriver]
        test_env: Dict[str, Any]
        callback_name: str
        slot: int

    @dataclass(frozen=True)
    class Output:
        """Subprocess output data."""

        result_queue: ResultQueue

    exchange_file: str
    """Name of the file to exchange data with the subprocess."""

    def __init__(
        self,
        uid: str,
        driver: TestDriver,
        callback_name: str,
        slot: int,
        running_status: RunningStatus,
        env: e3.env.Env,
    ):
        """Initialize a ProcessTestFragment.

        :param uid: UID of the test fragment (should be unique).
        :param driver: A TestDriver instance.
        :param callback_name: Name of the `driver` method to be executed by the
            job.
        :param slot: Slot ID allocated for the execution of this test fragment.
        :param running_status: See BaseTestFragment.running_status.
        :param env: Testsuite environment.
        """
        self.running_status = running_status
        self.result_queue = []
        self.skipped_for_abortion = False
        super().__init__(uid, driver, callback_name, slot, env)

    def clear_driver_data(self) -> None:
        joker: Any = None
        self.driver = joker

    def start(self) -> Run | None:
        # The decision to start this fragment or not (i.e. the call to
        # must_run) is taken here: it would be wrong to call it later, since
        # testsuite may not be aborted right now but may be aborted while this
        # fragment runs. Once a fragment has started, we always wait for its
        # completion.
        if not self.must_run():
            self.skipped_for_abortion = True
            self.return_value = None
            return None

        self.running_status.start(self)

        # Create the exchange file: write the test environment to it
        worker_input = self.Input(
            self.uid,
            type(self.driver),
            self.driver.test_env,
            self.callback_name,
            self.slot,
        )
        with tempfile.NamedTemporaryFile(
            prefix=self.uid + "-",
            dir=self.env.exchange_dir,
            delete=False,
        ) as f:
            self.exchange_file = f.name
            pickle.dump(worker_input, f)

        # Run the worker process, checking that it runs to completion
        # successfully.
        return Run(
            cmds=[
                "e3-run-test-fragment",
                self.env.env_filename,
                f.name,
            ],
            output=PIPE,
            error=STDOUT,
            input=DEVNULL,
            bg=True,
        )

    def poll(
        self,
        scheduler: e3.testsuite.multiprocess_scheduler.MultiprocessScheduler,
    ) -> bool:
        # If this fragment was not run, there is no subprocess to wait: the
        # caller can act as if the fragment had completed.
        if self.skipped_for_abortion:
            return False
        else:
            return super().poll(scheduler)

    def collect_result(self) -> None:
        # If this fragment was not run, there is no result to collect
        if self.skipped_for_abortion:
            return

        # Let extract_result_queue put results in the driver's result queue,
        # then forward them to this fragment's.
        self.driver.result_queue = []
        self.extract_result_queue()
        self.result_queue = self.driver.result_queue

        # Now cleanup temporary files and declare this fragment as completed
        rm(self.exchange_file)
        self.running_status.complete(self)

    def extract_result_queue(self) -> None:
        """Read the result queue from the exchange file.

        Try to extract the result queue from the exchange file and put results
        in the driver's result queue. If anything goes sour, create an error
        result in the same result queue.
        """
        # If this fragment was not run, collect_result should not call this
        assert not self.skipped_for_abortion

        # Make sure the process exitted with no error
        if self.process.status != 0:
            result = TestResult(
                f"{self.uid}__internalerror",
                env=self.driver.test_env,
                status=TestStatus.ERROR,
                msg="Worker process stopped with an error",
            )
            assert self.process.out is not None
            result.log += self.process.out
            self.driver.push_result(result)
            return

        # Now read the output from the exchange file. If there is anything
        # unexpected, give up decoding the actual results and create an error
        # result instead.
        try:
            with open(self.exchange_file, "rb") as f:
                output = pickle.load(f)
                assert isinstance(output, self.Output)
        except Exception as e:
            self.push_error_result(e)
            return

        self.driver.result_queue.extend(output.result_queue)


# TODO: Investigate why the pytest-cov plugin fails to track coverage in the
# following function (it is supposed to handle subprocesses).


def run_fragment(argv: Optional[List[str]] = None) -> None:  # no cover
    """Run a test fragment.

    This function is meant to be the entry point of a standalone script, to run
    a fragment in a subprocess, separate from the main testsuite process.
    """
    from e3.testsuite import TestAbort

    args_parser = argparse.ArgumentParser(
        description="Internal script for e3-testsuite to run a test fragment"
        " in a subprocess."
    )
    args_parser.add_argument(
        "env-filename", help="Name of the file to load with e3.env.Env.restore"
    )
    args_parser.add_argument(
        "exchange-file",
        help="Name of the file that contains test driver data and in which to"
        " write test results.",
    )
    args = args_parser.parse_args(argv)
    env_filename = getattr(args, "env-filename")
    exchange_file = getattr(args, "exchange-file")

    # Load the testsuite environment
    env = e3.env.Env()
    env.restore(env_filename)

    # Give access to modules that the testsuite main has access to
    sys.path = env.modules_search_path + sys.path

    # Load data needed to run the test fragment
    with open(exchange_file, "rb") as f:
        worker_input = pickle.load(f)
        assert isinstance(worker_input, ProcessTestFragment.Input)

    # Run the test fragment, silently ignoring TestAbort exceptions
    driver = worker_input.driver_cls(env, worker_input.test_env)
    callback = getattr(driver, worker_input.callback_name)
    try:
        callback(prev={}, slot=worker_input.slot)
    except TestAbort:
        pass
    except Exception:
        TestFragment.static_push_error_result(
            worker_input.fragment_uid, driver
        )

    # Forward test results to the testsuite
    worker_output = ProcessTestFragment.Output(driver.result_queue)
    with open(exchange_file, "wb") as f:
        pickle.dump(worker_output, f)



import logging
import tempfile

import e3.log
import yaml
from e3.env import Env
from e3.fs import rm
from e3.os.fs import cd, which

import pytest

# Activate full debug logs
e3.log.activate(level=logging.DEBUG, e3_debug=True)


class RequirementCoverage(object):
    """Track requirements <-> tests."""

    output_filename = None
    results = {}

    @classmethod
    def dump(cls):
        if cls.output_filename:
            with open(cls.output_filename, "w") as f:
                yaml.dump(cls.results, f)


@pytest.fixture(autouse=True)
def env_protect(request):
    """Protection against environment change.

    The fixture is enabled for all tests and does the following:

    * store/restore env between each tests
    * create a temporary directory and do a cd to it before each
      test. The directory is automatically removed when test ends
    """
    Env().store()
    tempd = tempfile.mkdtemp()
    cd(tempd)

    def restore_env():
        Env().restore()
        rm(tempd, True)

    request.addfinalizer(restore_env)


def pytest_addoption(parser):
    parser.addoption(
        "--ci", action="store_true", help="Tests are running on a CI server"
    )
    parser.addoption(
        "--requirement-coverage-report", help="Report requirement coverage"
    )


@pytest.fixture(autouse=True)
def require_git(request):
    marker = request.node.get_closest_marker("git")
    if marker:
        git(request)


@pytest.fixture
def git(request):
    if not which("git"):
        if request.config.getoption("ci"):
            pytest.fail("git not available")
        else:
            pytest.skip("git not available")


def pytest_configure(config):
    RequirementCoverage.output_filename = config.getoption(
        "requirement_coverage_report"
    )


def pytest_itemcollected(item):
    """Keep track of all test linked to a requirement."""
    doc = item.obj.__doc__
    if RequirementCoverage.output_filename and doc:
        for line in item.obj.__doc__.splitlines():
            line = line.strip().strip(".")
            if line.startswith("REQ-"):
                RequirementCoverage.results[item.obj.__name__] = line


def pytest_collectreport(report):
    """Output requirement coverage report."""
    RequirementCoverage.dump()



"""Helpers for testcases."""

import os.path

from e3.testsuite import Testsuite
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import Log, TestResult as Result
from e3.testsuite.testcase_finder import ParsedTest


class MultiSchedulingSuite(Testsuite):
    """Helper to manually select the scheduler when running the testsuite."""

    def add_options(self, parser):
        parser.add_argument("--multiprocessing", action="store_true")

    def compute_use_multiprocessing(self):
        return self.main.args.multiprocessing


def create_testsuite(
    test_names,
    driver_cls,
    ts_cls=Testsuite,
    adjust_dag_deps=None,
):
    """Create a helper testsuite class.

    The point of this function is to provide a simple mean to create a
    testsuite & its tests with no filesystem support.

    :param test_names: List of names for tests to run.
    :param driver_cls: Test driver to use for these tests.
    :param ts_cls: Testsuite base class.
    :param adjust_dag_deps: Optional callback to adjust DAG dependencies.
    """
    tests = [
        ParsedTest(
            test_name=test_name,
            driver_cls=driver_cls,
            test_env={},
            test_dir=".",
            test_matcher=None,
        )
        for test_name in test_names
    ]

    class MySuite(ts_cls):
        def get_test_list(self, sublist):
            return tests

        def adjust_dag_dependencies(self, dag):
            if adjust_dag_deps:
                adjust_dag_deps(self, dag)

    return MySuite


def check_result_dirs(new=None, old=None, new_dir=None, old_dir=None):
    """Check the content of testsuite result directories.

    :param new: Mapping from test names to expected test statuses for
        "new_dir".
    :param old: Likewise, but for "old_dir".
    :param new_dir: Directory that contains new test results. If left to None,
        use "out/new" in the current directory.
    :param old_dir: Likewise, for old test results. If left to None, use
        "out/old" in the current directory.
    """
    new = new if new is not None else {}
    old = old if old is not None else {}
    dirs = {
        "new": new_dir or os.path.join("out", "new"),
        "old": old_dir or os.path.join("out", "old"),
    }
    expected_data = {"new": new, "old": old}
    actual_data = {"new": {}, "old": {}}

    for d in ("new", "old"):
        output_dir = dirs[d]
        if os.path.exists(os.path.join(output_dir, "_index.json")):
            index = ReportIndex.read(dirs[d])
            for e in index.entries.values():
                actual_data[d][e.test_name] = e.status
    assert expected_data == actual_data, f"{expected_data} != {actual_data}"


def check_result_from_prefix(suite, prefix, status, msg):
    """Check the content of a result from its name prefix."""
    matches = []
    for key, value in suite.report_index.entries.items():
        if key.startswith(prefix):
            matches.append(value)
    assert len(matches) == 1, "Exactly one entry matching expected"
    result = matches[0]
    assert result.status == status
    assert result.msg == msg


def check_result(suite, test_name, status, msg):
    """Check the status and message for a test result."""
    result = suite.report_index.entries[test_name]
    assert result.status == status
    assert result.msg == msg


def extract_results(testsuite):
    """Extract synthetic test results from a testsuite run."""
    return {
        e.test_name: e.status for e in testsuite.report_index.entries.values()
    }


def run_testsuite_status(cls, args=None, multiprocessing=False):
    """Instantiate a Testsuite subclass, run it and return it and its sttus."""
    args = list(args) if args is not None else []
    if multiprocessing:
        args.append("--multiprocessing")
    suite = cls()
    return (suite, suite.testsuite_main(args))


def run_testsuite(cls, args=None, multiprocessing=False, expect_failure=False):
    """Instantiate a Testsuite subclass and run it."""
    args = args if args is not None else []
    suite, status = run_testsuite_status(cls, args, multiprocessing)
    if expect_failure:
        assert status != 0
    else:
        assert status == 0
    return suite


def suite_logs(caplog):
    """Extract messages of testsuite log records."""
    return {r.getMessage() for r in caplog.records if r.name == "testsuite"}


def create_report(results, tmp_path):
    """Create a report index in "tmp_path" for the given results."""
    index = ReportIndex(tmp_path)
    for r in results:
        index.add_result(r.summary, r.save(tmp_path))
    index.write()
    return index


def create_result(
    name,
    status,
    msg="",
    log="",
    out=None,
    expected=None,
    diff=None,
    time=None,
    failure_reasons=None,
    encoding=None,
):
    """Create a TestResult instance."""
    result = Result(name, status=status, msg=msg)
    result.env = {}
    result.log += log
    if out is not None:
        result.out = Log(out)
    if expected is not None:
        result.expected = Log(expected)
    if diff is not None:
        result.diff = Log(diff)
    result.time = time
    if failure_reasons:
        result.failure_reasons.update(failure_reasons)

    if encoding:
        result.env["encoding"] = encoding

    return result



"""Tests for the results output facilities."""

import os.path
import shutil

from e3.testsuite import Testsuite as Suite
from e3.testsuite.driver import TestDriver as Driver
from e3.testsuite.result import TestStatus as Status

from .utils import check_result_dirs, run_testsuite


# Helpers for all tests in this module


class MyDriver(Driver):
    def add_test(self, dag):
        self.add_fragment(dag, "run")

    def run(self, prev, slot):
        self.result.set_status(self.env.return_status)
        self.push_result()


class Mysuite(Suite):
    tests_subdir = "simple-tests"
    test_driver_map = {"default": MyDriver}
    default_driver = "default"

    def add_options(self, parser):
        parser.add_argument("return_status")

    def set_up(self):
        self.env.return_status = Status[self.main.args.return_status]


results_pass = {"test1": Status.PASS, "test2": Status.PASS}
results_fail = {"test1": Status.FAIL, "test2": Status.FAIL}
results_skip = {"test1": Status.SKIP, "test2": Status.SKIP}


def run(status, args=None, expect_failure=False):
    args = args if args is not None else []
    run_testsuite(
        Mysuite, args=args + [status.name], expect_failure=expect_failure
    )


# Actual tests


def test_default():
    """Check that with default options, results only get overwritten."""
    # Do a first testsuite run, checking the results
    run(Status.PASS)
    check_result_dirs(new=results_pass)
    assert not os.path.exists(os.path.join("out", "old"))

    # Then do a second one. We expect the "new" directory to just get replaced,
    # and still have no "old" directory.
    run(Status.FAIL, expect_failure=True)
    check_result_dirs(new=results_fail)
    assert not os.path.exists(os.path.join("out", "old"))


def test_output_dir(tmp_path):
    """Check that --output-dir alone behaves as expected."""
    args = ["--output-dir", str(tmp_path)]
    new_dir = str(tmp_path / "new")
    old_path = tmp_path / "old"

    # Do a first testsuite run, checking the results
    run(Status.PASS, args)
    check_result_dirs(new=results_pass, new_dir=new_dir)
    assert not old_path.exists()

    # Run it a second time. We expect the new results to replace the previous
    # ones in the "new" directory and still not to have any old results.
    run(Status.FAIL, args, expect_failure=True)
    check_result_dirs(new=results_fail, new_dir=new_dir)
    assert not old_path.exists()


def test_rotation(tmp_path):
    """Check that results rotation works as expected."""

    def do_reset():
        shutil.rmtree(str(tmp_path))
        tmp_path.mkdir()

    def check(args, new_dir, old_dir, reset=True):
        if reset:
            do_reset()

        # Do a first testsuite run, checking the results
        run(Status.PASS, args)
        check_result_dirs(new=results_pass, new_dir=new_dir)
        assert not os.path.exists(old_dir)

        # Run it a second time. We expect the previous "new" directory to be
        # renamed to "old", and the new results to be written to a fresh "new"
        # directory.
        run(Status.FAIL, args, expect_failure=True)
        check_result_dirs(
            new=results_fail,
            new_dir=new_dir,
            old=results_pass,
            old_dir=old_dir,
        )

        # Run it a third time. We expect:
        #
        # * the previous "old" directory to be removed
        # * the previous "new" directory to be renamed to "old"
        # * the new results to be written to a fresh "new" directory.
        run(Status.SKIP, args)
        check_result_dirs(
            new=results_skip,
            new_dir=new_dir,
            old=results_fail,
            old_dir=old_dir,
        )

    # Check log rotation with...

    # ... default output directories (out/new and out/old)
    saved_cwd = os.getcwd()
    os.chdir(str(tmp_path))
    try:
        check(
            args=["--rotate-output-dirs"],
            new_dir=str(tmp_path / "out" / "new"),
            old_dir=str(tmp_path / "out" / "old"),
            reset=False,
        )
    finally:
        os.chdir(saved_cwd)

    # ... output directory only (new and old are subdirectories)
    check(
        args=["--rotate-output-dirs", "--output-dir", str(tmp_path / "foo")],
        new_dir=str(tmp_path / "foo" / "new"),
        old_dir=str(tmp_path / "foo" / "old"),
    )

    # ... both explicit output directory and old output directory
    check(
        args=[
            "--rotate-output-dirs",
            "--output-dir",
            str(tmp_path / "foo"),
            "--old-output-dir",
            str(tmp_path / "bar"),
        ],
        new_dir=str(tmp_path / "foo"),
        old_dir=str(tmp_path / "bar"),
    )


def test_empty_old_output_dir_report(tmp_path):
    """Check text report generation with an empty old output dir.

    The testsuite framework used to crash in such a case.
    """
    new_dir = tmp_path / "new"
    old_path = tmp_path / "old"
    old_path.mkdir()

    args = [
        f"--output-dir={new_dir}",
        f"--old-output-dir={old_path}",
        "--generate-text-report",
    ]
    run(Status.PASS, args)



"""Tests for the multiprocessing scheduler for test fragments."""

import sys

from e3.testsuite.driver import BasicTestDriver as BasicDriver
from e3.testsuite.result import TestStatus as Status

from .utils import create_testsuite, extract_results, run_testsuite


class TestEnable:
    """Test that automatic enabling of multiprocessing works as expected."""

    ENABLED_MSG = "multiprocessing enabled"
    DISABLED_MSG = "multiprocessing disabled"

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(
                Status.PASS,
                (
                    TestEnable.ENABLED_MSG
                    if self.env.use_multiprocessing
                    else TestEnable.DISABLED_MSG
                ),
            )
            self.push_result()

    def run_testsuite(self, enable_expected, args=None, supported=None):
        """Run a testsuite with a single test with MyDriver."""
        MySuite = create_testsuite(["single_test"], TestEnable.MyDriver)
        if supported is not None:
            MySuite.multiprocessing_supported = supported

        suite = run_testsuite(MySuite, args=args)
        assert len(suite.report_index.entries) == 1
        result = suite.report_index.entries["single_test"]
        assert result.status == Status.PASS
        if enable_expected:
            assert result.msg == self.ENABLED_MSG
        else:
            assert result.msg == self.DISABLED_MSG

    def test_parallel(self):
        """Test that multiprocessing is enabled when support."""
        self.run_testsuite(True, ["-j17"], supported=True)

    def test_not_supported(self):
        """Test that multiprocessing is not enabled when not supported."""
        self.run_testsuite(False, ["-j17"], supported=False)

    def test_too_few_jobs(self):
        """Test that multiprocessing is disabled when parallelism is low."""
        self.run_testsuite(False, ["-j16"], supported=True)

    def test_force(self):
        """Check that --force-multiprocessing have precedence."""
        self.run_testsuite(
            True, ["-j1", "--force-multiprocessing"], supported=False
        )


class TestCollectError:
    """Check issues in result collection are properly handled.

    * check_process_error checks that non-zero subprocess exit codes are
      reported.
    * check_unpicklable checks that corrupted process outputs are reported.
    """

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            method = getattr(self, self.test_name)
            method()

        def process_error(self):
            sys.exit(1)

        def unpicklable(self):
            # Exit "properly" with an empty (and thus invalid) exchange file
            exchange_file = sys.argv[-1]
            with open(exchange_file, "wb"):
                pass
            sys.exit(0)

    def run_testsuite(self, test_name):
        MySuite = create_testsuite([test_name], TestCollectError.MyDriver)
        results = extract_results(
            run_testsuite(
                MySuite,
                args=["--force-multiprocessing", "-E"],
                expect_failure=True,
            )
        )
        keys = sorted(results)
        assert len(keys) == 1
        return keys[0], results[keys[0]]

    def test_process_error(self):
        """Check that non-zero subprocess exit codes are reported."""
        key, result = self.run_testsuite("process_error")
        assert key.startswith("process_error.analyze__internalerror")
        assert result == Status.ERROR

    def test_unpicklable(self):
        key, result = self.run_testsuite("unpicklable")
        assert key == "unpicklable.analyze"
        assert result == Status.ERROR



from __future__ import annotations

import os.path
import re
import sys
import time
from typing import Dict, List, Optional, Pattern, Tuple

from e3.fs import cp, sync_tree
from e3.testsuite.driver.classic import ProcessResult, TestAbortWithError
from e3.testsuite.driver.diff import (
    DiffTestDriver,
    LineByLine,
    OutputRefiner,
    PatternSubstitute,
)
from e3.testsuite.control import AdaCoreLegacyTestControlCreator
from e3.testsuite.result import FailureReason


class AdaCoreLegacyTestDriver(DiffTestDriver):
    """Test driver for legacy AdaCore testsuites.

    This test driver expects a "test.cmd", "test.sh" or "test.py" script in its
    directory, runs it, and succeeds iff its output matches the content of the
    "test.out" file (or is empty, if that file is missing). Execution of tests
    is controlled by "test.opt" files.

    This driver assumes that the list of discriminants for the "test.opt" file
    is available as "self.env.discs" (i.e. it is a list of strings), and that
    the environ for subprocesses is available as "self.env.test_environ".

    If the TEST_SUPPORT_DIR environment variable is defined, consider that it
    contains the name of a directory that contains a "support.sh" script, that
    is sourced at the beginning of every CMD/Shell test script.
    """

    # We have special procedures to copy test material
    copy_test_directory = False

    # Work on raw bytes: some tests work with ISO-8859-1, others UTF-8, others
    # JIS, ... and we don't have the metadata to know which.
    default_encoding = "binary"

    # By default, ignore white characters in outputs
    diff_ignore_white_chars = True

    argv: List[str]
    test_environ: Dict[str, str]

    test_process: Optional[ProcessResult]

    @property
    def test_control_creator(self) -> AdaCoreLegacyTestControlCreator:
        assert isinstance(self.env.discs, list)
        return AdaCoreLegacyTestControlCreator(self.env.discs)

    @property
    def baseline_file(self) -> Tuple[str, bool]:
        # the "_baseline_file" attribute is defined during the "set_up" stage,
        # and our baseline is never a regexp.
        return (self._baseline_file, False)

    def set_up(self) -> None:
        super().set_up()
        assert self.test_control.opt_results is not None

        # If we have a non-standard output baseline file, make sure it is
        # present.
        baseline = self.test_control.opt_results["OUT"]
        self._original_baseline_file = self.test_dir(baseline)
        if not baseline.endswith("test.out") and not os.path.isfile(
            self._original_baseline_file
        ):
            raise TestAbortWithError(
                "cannot find output file {}".format(baseline)
            )

        # Copy test material to $working_dir/src and make sure we have a
        # baseline file.
        sync_tree(self.test_dir(), self.working_dir("src"), delete=True)
        self._baseline_file = self.working_dir("src", baseline)
        if not os.path.isfile(self._baseline_file):
            with open(self._baseline_file, "w"):
                pass

        # Prepare the test script execution
        self.timeout = int(self.test_control.opt_results["RLIMIT"])
        assert isinstance(self.env.test_environ, dict)
        self.test_environ = dict(self.env.test_environ)
        self.argv = self.get_script_command_line()
        self.test_process = None

    default_substitutions: List[Tuple[Pattern[str], str]] = [
        # Remove ".exe" suffix for output files. This will for instance turn
        # "gcc -o main.exe main.adb" into "gcc -o main main.adb".
        (re.compile(r"-o(.*).exe"), r"-o \1"),
        # Convert references to environment variables: "%VAR%" -> "$VAR"
        (re.compile(r"%([^ ]*)%"), r'"$\1"'),
        # TODO??? Imported as-is from gnatpython.testdriver. It's not clear
        # what this tries to do given that escape characters are interpreted
        # literally ('r' string prefix).
        (re.compile(r"(\032|\015)"), r""),
        # Replace environment variable definitions:
        # "set FOO = BAR" -> 'FOO="BAR"; export FOO'.
        (re.compile(r"set *([^ =]+) *= *([^ ]*)"), r'\1="\2"; export \1'),
    ]

    @property
    def cmd_substitutions(self) -> List[Tuple[Pattern[str], str]]:
        """
        List of substitutions to apply to scripts.

        This returns a list of patterns/replacements couples for substitutions
        to apply to scripts in order to convert them from "cmd" syntax to
        Bourne shell.
        """
        return list(self.default_substitutions)

    @property
    def script_encoding(self) -> str:
        """Return the encoding to decode shell scripts.

        By default, this is the same as the ``default_encoding`` property,
        except in one case: when it returns ``binary``: assume UTF-8 in that
        case.
        """
        result = self.default_encoding
        return "utf-8" if result == "binary" else result

    def get_script_command_line(self) -> List[str]:
        """Return the command line to run the test script."""
        # Command line computation depends on the kind of script (Python or
        # shell).
        assert isinstance(self.env.discs, list)

        # Make sure the test script is present in the working directory
        assert self.test_control.opt_results is not None
        script_filename = self.test_control.opt_results["CMD"]
        self.script_file = self.working_dir("src", script_filename)
        if not os.path.isfile(self.script_file):
            raise TestAbortWithError(
                "cannot find script file {}".format(script_filename)
            )

        _, ext = os.path.splitext(self.script_file)

        # Some tests have a _suffix in their extension. Using .startwith
        # ensures we don't treat ".cmd_x86" as ".sh".
        is_cmd = ext.startswith(".cmd")
        must_convert = is_cmd and (
            self.env.host.os.name != "windows" or "FORCE_SH" in self.env.discs
        )

        if ext == ".py":
            return [sys.executable, self.script_file]

        elif not is_cmd or must_convert:
            # If running a Bourne shell script, not running on Windows, or if
            # specifically asked to use a Bourne shell, create a shell script
            # to run instead of the given test script.
            new_script = []

            # First, make sure the current directory is in the PATH, to ease
            # running just-built programs in test scripts.
            new_script.append("PATH=.:$PATH; export PATH")

            # TODO: filesize_limit handling

            # If "self.env.support_dir" designates a directory that contains a
            # "support.sh" file, make the test script source it.
            support_dir = os.environ.get("TEST_SUPPORT_DIR", "")
            support_script = os.path.join(support_dir, "support.sh")
            if support_dir and os.path.isfile(support_script):
                new_script.append(". $TEST_SUPPORT_DIR/support.sh")

            # Read all lines in the original test script
            script_encoding = self.script_encoding
            with open(self.script_file, encoding=script_encoding) as f:
                # Get rid of potential whitespaces and CR at the end of
                # each line.
                for line in f:
                    line = line.rstrip()
                    if must_convert:
                        # convert the "cmd" syntax to Bourne shell
                        for pattern, replacement in self.cmd_substitutions:
                            line = pattern.sub(replacement, line)
                    new_script.append(line)

            # Write the shell script and schedule its execution with "bash". On
            # Windows, Python interpreters may automatically add CR bytes
            # before LR ones: open the file in binary mode to avoid this
            # behavior, as "bash" would complain about CR bytes.
            new_script_filename = self.working_dir("__test.sh")
            with open(new_script_filename, "wb") as f:
                for line in new_script:
                    f.write(line.encode(script_encoding))
                    f.write(b"\n")
            return ["bash", new_script_filename]

        else:  # os-specific
            # We are running on Windows, so we can use "cmd" to run the
            # original script. Just make sure it has the correct extension.
            script_file = self.script_file
            if not script_file.endswith(".cmd"):
                script_file = self.working_dir("test__.cmd")
                cp(self.script_file, script_file)
            return ["cmd.exe", "/q", "/c", script_file]

    @property
    def output_refiners(self) -> List[OutputRefiner[bytes]]:
        return [
            # Remove platform specificities in relative filenames
            PatternSubstitute(rb"\\", rb"/"),
            # Remove ".exe" extension and CR characters anywhere in outputs.
            # TODO: same question as in the TODO in "cmd_substitutions".
            PatternSubstitute(rb"(\.exe\b|\015)", rb""),
            # Remove occurences of the "src" working dir subdirectory
            LineByLine(
                PatternSubstitute(
                    rb"[^ '\"]*"
                    + os.path.basename(self.working_dir()).encode("ascii")
                    + rb"/src/",
                    rb"",
                )
            ),
        ]

    def run(self) -> None:
        # Run the test script and record execution time. Note that the
        # status code is not significant (catch_error=False).
        start_time = time.time()
        self.test_process = self.shell(
            args=self.argv,
            cwd=self.working_dir("src"),
            env=self.test_environ,
            timeout=self.timeout,
            catch_error=False,
        )
        self.result.time = time.time() - start_time

    def compute_failures(self) -> List[str]:
        # First, do compute the failures and let the baseline rewriting
        # machinery do its magic on the baseline it is given (i.e. the copy in
        # the working directory: see set_up).
        result = super().compute_failures()

        # Now, make sure we propagate the new baseline to the test directory
        if not self.test_control.xfail and getattr(
            self.env, "rewrite_baselines", False
        ):
            with open(self._baseline_file, "rb") as f:
                content = f.read()

            # Materialize empty baselines as missing baseline file, but only
            # for the default baseline file.
            default_baseline_file = self.test_dir("test.out")
            if (
                content
                or self._original_baseline_file != default_baseline_file
            ):
                with open(self._original_baseline_file, "wb") as f:
                    f.write(content)
            elif os.path.exists(default_baseline_file):
                os.remove(default_baseline_file)

        if self.test_process is not None and self.process_may_have_timed_out(
            self.test_process
        ):
            self.result.failure_reasons.add(FailureReason.TIMEOUT)
            result.append("test timed out")

        return result



"""Tests for the XUnit report feature."""

import os.path
import xml.etree.ElementTree as ET

import yaml

from e3.fs import mkdir
from e3.testsuite import Testsuite as Suite
from e3.testsuite.driver import TestDriver as Driver
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.report.xunit import (
    XUnitImporter,
    XUnitImporterApp,
    convert_main,
)
from e3.testsuite.result import TestStatus as Status

from .utils import create_testsuite, run_testsuite


def write_xfails_yaml(filename, xfails):
    """Write a XFAILs YAML file."""
    with open(filename, "w") as f:
        yaml.dump(xfails, f)


class TestBasic:
    """Check that requesting a XUnit testsuite report works."""

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.log += "Work is being done..."

            if self.test_env["test_name"] == "test1":
                self.result.set_status(Status.PASS, "all good")
            else:
                self.result.set_status(Status.FAIL, "test always fail!")
            self.push_result()

    def test(self):
        class Mysuite(Suite):
            tests_subdir = "simple-tests"
            test_driver_map = {"default": self.MyDriver}
            default_driver = "default"

        xunit_file = "xunit.xml"
        run_testsuite(
            Mysuite, ["--xunit-output", xunit_file], expect_failure=True
        )

        # For now, just check that this produces a valid XML file
        ET.parse(xunit_file)


class TestXFail:
    """Check that XFAIL tests are reported as skipped."""

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.set_status(Status.XFAIL)
            self.push_result()

    def test(self, tmp_path):
        xunit_file = str(tmp_path / "xunit.xml")
        run_testsuite(
            create_testsuite(["mytest"], self.MyDriver),
            ["--xunit-output", xunit_file],
        )

        ts = ET.parse(xunit_file).getroot()
        assert ts.attrib["errors"] == "0"
        assert ts.attrib["failures"] == "0"
        assert ts.attrib["skipped"] == "1"
        assert ts.attrib["tests"] == "1"


class TestTime:
    """Check that testsuite and testcase durations are properly reported."""

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.time = 1.2345
            self.result.set_status(Status.PASS)
            self.push_result()

    def test(self, tmp_path):
        xunit_file = str(tmp_path / "xunit.xml")
        run_testsuite(
            create_testsuite(["mytest"], self.MyDriver),
            ["--xunit-output", xunit_file],
        )

        testsuites = ET.parse(xunit_file).getroot()
        testsuite = testsuites[0]
        testcase = testsuite[0]

        # Sanity checks
        assert testsuites.tag == "testsuites"
        assert testsuite.tag == "testsuite"
        assert testcase.tag == "testcase"
        assert testcase.get("name") == "mytest"

        # Check that we have time attributes for both the <testsuites> and
        # <testsuite> tags. Since we do not control the time it takes to
        # actually run it, we can just check for the presence of these
        # attributes, not their actual values.
        assert isinstance(testsuites.get("time"), str)
        assert testsuites.get("time") == testsuite.get("time")

        # We do control the value for the testcase
        assert testcase.get("time") == "1.234"


class TestControlChars:
    """Check XUnit reports with control characters.

    Check that XUnit reports are valid when control chars are present in test
    outputs.
    """

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.log += "Control character: \x01\nDone.\n"
            self.result.set_status(Status.FAIL, "Another control char: \x02")
            self.push_result()

    def test(self, tmp_path):
        xunit_file = str(tmp_path / "xunit.xml")
        run_testsuite(
            create_testsuite(["mytest"], self.MyDriver),
            ["--xunit-output", xunit_file],
            expect_failure=True,
        )
        testsuites = ET.parse(xunit_file).getroot()
        testsuite = testsuites[0]
        testcase = testsuite[0]
        failure = testcase[0]

        # Sanity checks
        assert testsuites.tag == "testsuites"
        assert testsuite.tag == "testsuite"
        assert testcase.tag == "testcase"
        assert testcase.get("name") == "mytest"
        assert failure.tag == "failure"

        assert failure.get("message") == "Another control char: \\x02"
        assert failure.text == "Control character: \\x01\nDone.\n"


def test_import(tmp_path, capsys):
    """Test that the xUnit importer works as expected."""
    xml_filename = str(tmp_path / "tmp.xml")
    with open(xml_filename, "w") as f:
        f.write(
            """<?xml version="1.0" encoding="utf-8"?>
            <testsuites name="MyTestsuites">
                <testsuite name="Normal">

                    <!-- Elements discarded during the import. -->
                    <properties>
                        <property name="to-ignore"></property>
                    </properties>
                    <system-out>
                        To ignore
                    </system-out>
                    <system-err>
                        To ignore
                    </system-err>

                    <!-- Test import of the time attribute. -->
                    <testcase name="test1" time="0.1"></testcase>

                    <!-- Test slugification. -->
                    <testcase name="Test2-\xe9"></testcase>
                    <testcase name="Test2-#$^"></testcase>

                    <!-- Test de-duplication of test names. -->
                    <testcase name="Test2-e"></testcase>
                    <testcase name="Test2-e"></testcase>

                    <!-- Test the various test statuses. -->
                    <testcase name="test-failure">
                        <failure>Some failure logging</failure>
                    </testcase>

                    <testcase name="test-error">
                        <error>Some error logging</error>
                    </testcase>

                    <testcase name="test-skipped">
                        <skipped>Some skip logging</skipped>
                    </testcase>

                    <!-- Test the import of status message. -->
                    <testcase name="test-failure-message">
                        <failure message="Some failure message">\
Some failure logging</failure>
                    </testcase>

                    <!-- Test handling of classname. -->
                    <testcase name="test_name" classname="MyClass">
                    </testcase>

                    <!-- Test handling of multiline messages. -->
                    <testcase name="test-failure-multiline-message">
                        <failure message="Some multi&#10;
line&#10;
message...">\
Some failure logging</failure>
                    </testcase>

                    <!-- Test handling of too long messages. -->
                    <testcase name="test-failure-too-long-message">
                        <failure message="Some extremely very ultra long
message. Viewers may not like it, so we are going to strip it to ~200 colons.
The imported report will include only a small prefix of this too long message.
They will not see the complete message. This is okay, as in known cases where
messages are too long, their content is actually also present in the log, which
is not capped, so no content is lost in practice.
">Some failure logging</failure>
                    </testcase>

                </testsuite>

                <!-- Test usage of XFAILs. -->
                <testsuite name="XFails">
                    <testcase name="test-ok"></testcase>
                    <testcase name="test-failure">
                        <failure>Some failure logging</failure>
                    </testcase>
                    <testcase name="test-failure-message">
                        <failure message="Some failure message">\
Some failure logging</failure>
                    </testcase>
                    <testcase name="test-error">
                        <error>Some error logging</error>
                    </testcase>
                    <testcase name="test-skipped">
                        <skipped>Some skip logging</skipped>
                    </testcase>
                    <testcase name="pytest-skip">
                        <skipped type="pytest.skip">Some skip logging</skipped>
                    </testcase>
                    <testcase name="pytest-xfail">
                        <skipped
                          type="pytest.xfail"
                          message="Known bug">Some error logging</skipped>
                    </testcase>
                </testsuite>
            </testsuites>"""
        )

    xfails_filename = str(tmp_path / "xfails.yaml")
    write_xfails_yaml(
        xfails_filename,
        {
            "XFails.test-ok": "",
            "XFails.test-failure": "",
            "XFails.test-error": "",
            "XFails.test-skipped": "",
            "XFails.test-failure-message": "Expected failure message",
        },
    )

    results_dir = str(tmp_path / "results")
    convert_main(
        ["-o", results_dir, "--xfails", xfails_filename, xml_filename]
    )
    index = ReportIndex.read(results_dir)

    assert sorted(index.entries) == [
        "Normal.MyClass.test_name",
        "Normal.Test2",
        "Normal.Test2-e",
        "Normal.Test2-e.1",
        "Normal.Test2-e.2",
        "Normal.test-error",
        "Normal.test-failure",
        "Normal.test-failure-message",
        "Normal.test-failure-multiline-message",
        "Normal.test-failure-too-long-message",
        "Normal.test-skipped",
        "Normal.test1",
        "XFails.pytest-skip",
        "XFails.pytest-xfail",
        "XFails.test-error",
        "XFails.test-failure",
        "XFails.test-failure-message",
        "XFails.test-ok",
        "XFails.test-skipped",
    ]

    def check(test_name, status, message=None, time=None):
        e = index.entries[test_name]
        assert e.status == status
        assert e.msg == message
        if time is None:
            assert e.time is None
        else:
            assert str(e.time) == time

    def check_log(test_name, log):
        assert index.entries[test_name].load().log == log

    check("Normal.test1", Status.PASS, time="0.1")
    check("Normal.Test2", Status.PASS)
    check("Normal.Test2-e", Status.PASS)
    check("Normal.Test2-e.1", Status.PASS)
    check("Normal.Test2-e.2", Status.PASS)

    check("Normal.test-failure", Status.FAIL)
    check_log("Normal.test-failure", "Some failure logging")

    check(
        "Normal.test-failure-multiline-message",
        Status.FAIL,
        message="Some multi [...]",
    )
    check_log("Normal.test-failure-multiline-message", "Some failure logging")
    check(
        "Normal.test-failure-too-long-message",
        Status.FAIL,
        message=(
            "Some extremely very ultra long message. Viewers may not like it,"
            " so we are going to strip it to ~200 colons. The imported report"
            " will include only a small prefix of this too long message. They"
            " will no [...]"
        ),
    )
    check_log("Normal.test-failure-too-long-message", "Some failure logging")

    check("Normal.test-error", Status.ERROR)
    check_log("Normal.test-error", "Some error logging")

    check("Normal.test-skipped", Status.SKIP)
    check_log("Normal.test-skipped", "Some skip logging")

    check("Normal.test-failure-message", Status.FAIL, "Some failure message")
    check_log("Normal.test-failure-message", "Some failure logging")

    check("Normal.MyClass.test_name", Status.PASS)

    check("XFails.test-ok", Status.XPASS)

    check("XFails.test-failure", Status.XFAIL)
    check_log("XFails.test-failure", "Some failure logging")

    check("XFails.test-error", Status.XFAIL)
    check_log("XFails.test-error", "Some error logging")

    check("XFails.test-skipped", Status.SKIP)
    check_log("XFails.test-skipped", "Some skip logging")

    check(
        "XFails.test-failure-message",
        Status.XFAIL,
        "Some failure message (Expected failure message)",
    )
    check_log("XFails.test-failure-message", "Some failure logging")

    check("XFails.pytest-skip", Status.SKIP)
    check_log("XFails.pytest-skip", "Some skip logging")

    check("XFails.pytest-xfail", Status.XFAIL, "Known bug")
    check_log("XFails.pytest-xfail", "Some error logging")

    captured = capsys.readouterr()
    assert captured.out == ""

    # Ensure that stderr is empty apart from the logs that are
    # activated by the e3 pytest plugin
    assert not [
        line for line in captured.err.splitlines() if "DEBUG" not in line
    ]


def test_import_dirs(tmp_path):
    """Test XML reports search in the xUnit conversion script."""

    def write_xml(filename, testsuite_name, *test_names):
        mkdir(str(filename.parent))
        with open(str(filename), "w") as f:
            f.write(
                '<?xml version="1.0" encoding="utf-8"?>'
                '\n<testsuites name="MyTestsuites">'
                f'\n<testsuite name="{testsuite_name}">'
            )
            for n in test_names:
                f.write(f'\n<testcase name="{n}"></testcase>')
            f.write("\n</testsuite>\n</testsuites>")

    write_xml(tmp_path / "xml" / "f1.xml", "f1", "t1", "t2")
    write_xml(tmp_path / "xml" / "sdir" / "f2.xml", "f2", "t1")
    write_xml(tmp_path / "xml" / "sdir" / "ssdir" / "f3.xml", "f3", "t1")
    write_xml(tmp_path / "xml" / "odir" / "f4.xml", "f4", "t1")

    results_dir = str(tmp_path / "results")
    convert_main(
        [
            "-o",
            results_dir,
            str(tmp_path / "xml" / "f1.xml"),
            str(tmp_path / "xml" / "sdir"),
        ]
    )
    index = ReportIndex.read(results_dir)

    assert sorted(index.entries) == ["f1.t1", "f1.t2", "f2.t1", "f3.t1"]


def test_gaia(tmp_path):
    """Test GAIA-compatible report creation."""
    xml_report = str(tmp_path / "test.xml")
    with open(xml_report, "w") as f:
        f.write(
            """<?xml version="1.0" encoding="utf-8"?>
            <testsuites name="MyTestsuites">
              <testsuite name="MyTestsuite">
                <testcase name="MyTestcase"></testcase>
              </testsuite>
            </testsuites>
            """
        )

    results_dir = tmp_path / "results"
    convert_main(["--gaia-output", "-o", str(results_dir), xml_report])

    with open(str(results_dir / "results"), "r") as f:
        assert f.read() == "MyTestsuite.MyTestcase:OK:\n"


def test_dangling_xfails(tmp_path, capsys):
    """Test emission of warnings for dangling XFAILs."""
    xml_report = str(tmp_path / "test.xml")
    with open(xml_report, "w") as f:
        f.write(
            """<?xml version="1.0" encoding="utf-8"?>
            <testsuites name="MyTestsuites">
              <testsuite name="MyTestsuite">
                <testcase name="MyTestcase"></testcase>
              </testsuite>
            </testsuites>
            """
        )

    xfails_filename = str(tmp_path / "xfails.yaml")
    write_xfails_yaml(xfails_filename, {"foo": "reason", "bar": ""})

    convert_main(
        [
            "-o",
            str(tmp_path / "results"),
            "--xfails",
            xfails_filename,
            xml_report,
        ]
    )

    captured = capsys.readouterr()
    assert captured.out == (
        "warning: the following tests are expected to fail but are not present"
        " in the testsuite results:\n"
        "  bar\n"
        "  foo (reason)\n"
    )
    # Ensure that stderr is empty apart from the logs that are
    # activated by the e3 pytest plugin
    assert not [
        line for line in captured.err.splitlines() if "DEBUG" not in line
    ]


def test_app_override(tmp_path):
    """Test overriding in the XUnitImporterApp class."""
    xml_report = str(tmp_path / "test.xml")
    with open(xml_report, "w") as f:
        f.write(
            """<?xml version="1.0" encoding="utf-8"?>
            <testsuites name="TSList">
              <testsuite name="ts">
                <testcase name="tc1"></testcase>
                <testcase name="tc2">
                  <failure>Some logging</failure>
                </testcase>
              </testsuite>
            </testsuites>
            """
        )

    basic_results = {"ts.tc1": Status.PASS, "ts.tc2": Status.FAIL}

    def check_results(results_dir, expected):
        actual = {
            e.test_name: e.status
            for e in ReportIndex.read(results_dir).entries.values()
        }
        assert actual == expected

    # Check overriding the add_basic_options method/output_dir property.

    class CustomBasic(XUnitImporterApp):
        def add_basic_options(self, parser):
            parser.add_argument("e3-report-dir")

        @property
        def output_dir(self):
            return getattr(self.args, "e3-report-dir")

        def get_xfails(self):
            return {}

        def iter_xunit_files(self):
            yield xml_report

        @property
        def gaia_report_requested(self):
            return False

    CustomBasic().run(["custom-basic"])
    check_results("custom-basic", basic_results)

    # Check overriding the create_output_report_index method

    class CustomIndex(XUnitImporterApp):
        def create_output_report_index(self):
            mkdir("custom-index")
            return ReportIndex("custom-index")

    CustomIndex().run([xml_report])
    check_results("custom-index", basic_results)

    # Check overriding the add_options/get_xfails method

    class CustomXFails(XUnitImporterApp):
        def add_options(self, parser):
            parser.add_argument("--add-xfail", action="append")

        def get_xfails(self):
            result = {}
            for xfail in self.args.add_xfail:
                test_name, msg = xfail.split(":", 1)
                result[test_name] = msg
            return result

    CustomXFails().run(
        [xml_report, "-o", "custom-xfails", "--add-xfail=ts.tc2:foo"]
    )
    check_results(
        "custom-xfails", {"ts.tc1": Status.PASS, "ts.tc2": Status.XFAIL}
    )

    # Check overriding the create_importer method

    class MyXUnitImporter(XUnitImporter):
        def get_test_name(self, ts_name, tc_name, classname):
            return tc_name

    class CustomImporter(XUnitImporterApp):
        def create_importer(self):
            return MyXUnitImporter(self.index, self.xfails)

    CustomImporter().run([xml_report, "-o", "custom-importer"])
    check_results("custom-importer", {"tc1": Status.PASS, "tc2": Status.FAIL})

    # Check overriding the iter_xunit_files method

    class CustomIterXUnitFiles(XUnitImporterApp):
        def iter_xunit_files(self):
            yield xml_report

    CustomIterXUnitFiles().run(["-o", "custom-iter-xunit-files", "foo.xml"])
    check_results("custom-iter-xunit-files", basic_results)

    # Check overriding the gaia_report_requested property

    class CustomGAIA(XUnitImporterApp):
        @property
        def gaia_report_requested(self):
            return True

    CustomGAIA().run(["-o", "custom-gaia", xml_report])
    check_results("custom-gaia", basic_results)
    assert os.path.exists(os.path.join("custom-gaia", "results"))

    # Check overriding the tear_down method

    class CustomTearDown(XUnitImporterApp):
        def tear_down(self):
            with open(os.path.join(self.index.results_dir, "foo.txt"), "w"):
                pass

    CustomTearDown().run(["-o", "custom-tear-down", xml_report])
    check_results("custom-tear-down", basic_results)
    assert os.path.exists(os.path.join("custom-tear-down", "foo.txt"))



from __future__ import annotations

import os
import re

from pathlib import Path
import traceback
from typing import Any, Dict, IO, List, Optional, TYPE_CHECKING, Union

import e3.collection.dag
from e3.fs import rm, sync_tree
from e3.os.process import DEVNULL, PIPE, Run, STDOUT, quote_arg
from e3.testsuite import CleanupMode
from e3.testsuite.utils import DummyColors
from e3.testsuite.control import (
    TestControl,
    TestControlCreator,
    YAMLTestControlCreator,
)
from e3.testsuite.driver import TestDriver
from e3.testsuite.result import (
    Log,
    TestResult,
    TestStatus,
    binary_repr,
    truncated,
)
from e3.testsuite.utils import indent

from colorama import Fore, Style


if TYPE_CHECKING:
    from e3.os.process import DEVNULL_VALUE, PIPE_VALUE


class TestSkip(Exception):
    """
    Convenience exception to abort a testcase.

    When this exception is raised during test initialization or execution,
    consider that this testcase must be skipped (TestStatus.SKIP).
    """

    pass


class TestAbortWithError(Exception):
    """
    Convenience exception to abort a testcase.

    When this exception is raised during test initialization or execution,
    consider that something went wrong (TestStatus.ERROR).
    """

    pass


class TestAbortWithFailure(Exception):
    """Convenience exception to abort a testcase, considering it failed.

    When this exception is raised during test initialization or execution,
    consider that it failed (TestStatus.FAIL or TestStatus.XFAIL, depending on
    test control).
    """

    pass


# Regular expressions to match the "timeout" error message from the rlimit
# program.
TIMEOUT_OUTPUT_PATTERN = r"rlimit: Real time limit ([^\n]+) exceeded\n"
TIMEOUT_OUTPUT_STR_RE = re.compile(TIMEOUT_OUTPUT_PATTERN)
TIMEOUT_OUTPUT_BYTES_RE = re.compile(TIMEOUT_OUTPUT_PATTERN.encode("ascii"))


class ProcessResult:
    """Record results from a subprocess."""

    def __init__(self, status: int, out: Union[str, bytes]):
        """ProcessResult constructor.

        :param status: Process exit code.
        :param out: Captured process output stream.
        """
        self.status = status
        self.out = out


class ClassicTestDriver(TestDriver):
    """Enhanced test driver base class for common behaviors.

    This test driver provides several facilities to automate tasks that driver
    often duplicate in practice:

    * run subprocesses;
    * intercept subprocess failures and turn them into appropriate test
      statuses;
    * gather subprocess outputs to ``self.result.out``;
    * have support for automatic XFAIL/SKIP test results.
    """

    # These can also be colorama.Style and colorama.Fore, but we don't have
    # type hints for them.
    Style: DummyColors
    Fore: DummyColors

    # Depending on the default encoding, this can be either a log of strings or
    # a log of bytes.
    output: Log

    test_control: TestControl

    @property
    def copy_test_directory(self) -> bool:
        """
        Return whether to automatically copy test directory to working dir.

        If this returns True, the working directory is automatically
        synchronized to the test directory before running the testcase:
        """
        return True

    def run(self) -> None:
        """Run the testcase.

        Subclasses must override this.
        """
        raise NotImplementedError

    @property
    def default_process_timeout(self) -> int:
        """
        Return the default timeout for processes spawn in the ``shell`` method.

        The result is a number of seconds.
        """
        # Return the timeout defined in test.yaml, if present, otherwise return
        # our true default: 5 minutes.
        return self.test_env.get("timeout", 5 * 60)

    @property
    def default_encoding(self) -> str:
        """Return the default encoding to decode process outputs.

        If "binary", consider that process outputs are binary, so do not try to
        decode them to text.
        """
        return self.test_env.get("encoding", "utf-8")

    @property
    def test_control_creator(self) -> TestControlCreator:
        """Return a test control creator for this test.

        By default, this returns a YAMLTestControlCreator instance tied to this
        driver with an empty condition environment. Subclasses are free to
        override this to suit their needs: for instance returning a
        OptfileCreater to process "test.opt" files.
        """
        return YAMLTestControlCreator({})

    def shell(
        self,
        args: List[str],
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        catch_error: bool = True,
        analyze_output: bool = True,
        timeout: Optional[int] = None,
        encoding: Optional[str] = None,
        truncate_logs_threshold: Optional[int] = None,
        ignore_environ: bool = True,
        stdin: (
            DEVNULL_VALUE | PIPE_VALUE | str | bytes | Path | IO | None
        ) = DEVNULL,
    ) -> ProcessResult:
        """Run a subprocess.

        :param args: Arguments for the subprocess to run.
        :param cwd: Current working directory for the subprocess. By default
            (i.e. if None), use the test working directory.
        :param env: Environment to pass to the subprocess.
        :param catch_error: If True, consider that an error status code leads
            to a test failure. In that case, abort the testcase.
        :param analyze_output: If True, add the subprocess output to the
            ``self.output`` log.
        :param timeout: Timeout (in seconds) for the subprocess. Use
            ``self.default_timeout`` if left to None.
        :param encoding: Encoding to use when decoding the subprocess' output
            stream. If None, use the default enocding for this test
            (``self.default_encoding``, from the ``encoding`` entry in
            test.yaml).  If "binary", leave the output undecoded as a bytes
            string.
        :param truncate_logs_threshold: Threshold to truncate the subprocess
            output in ``self.result.log``. See
            ``e3.testsuite.result.truncated``'s ``line_count`` argument. If
            left to None, use the testsuite's ``--truncate-logs`` option.
        :param ignore_environ: Applies only when ``env`` is not None.
            When True (the default), pass exactly environment variables
            in ``env``. When False, pass a copy of ``os.environ`` that is
            augmented with variables in ``env``.
        :param stdin: Forwarded to ``e3.os.process.Run``'s ``input`` argument.
        """
        # By default, run the subprocess in the test working directory
        if cwd is None:
            cwd = self.test_env["working_dir"]

        if timeout is None:
            timeout = self.default_process_timeout

        if truncate_logs_threshold is None:
            truncate_logs_threshold = self.testsuite_options.truncate_logs

        # Run the subprocess and log it
        def format_header(label: str, value: Any) -> str:
            return "{}{}{}: {}{}\n".format(
                self.Style.RESET_ALL + self.Style.BRIGHT,
                label,
                self.Style.RESET_ALL,
                self.Style.DIM,
                value,
            )

        self.result.log += format_header(
            "Running",
            "{} (cwd={}{}{})".format(
                " ".join(quote_arg(a) for a in args),
                self.Style.RESET_ALL,
                cwd,
                self.Style.DIM,
            ),
        )

        process_info = {"cmd": args, "cwd": cwd}
        self.result.processes.append(process_info)

        subp = Run(
            cmds=args,
            cwd=cwd,
            output=PIPE,
            error=STDOUT,
            input=stdin,
            timeout=timeout,
            env=env,
            ignore_environ=ignore_environ,
        )

        # Testsuites sometimes need to deal with binary data (or unknown
        # encodings, which is equivalent), so always use subp.raw_out.
        stdout: Union[str, bytes]
        assert isinstance(subp.raw_out, bytes)
        stdout_bytes = subp.raw_out
        encoding = encoding or self.default_encoding
        if encoding != "binary":
            try:
                stdout = stdout_bytes.decode(encoding)
            except UnicodeDecodeError as exc:
                self.result.log += "Cannot decode subprocess output:\n\n"
                self.result.log += indent(binary_repr(stdout_bytes))
                raise TestAbortWithError(
                    "cannot decode process output ({}: {})".format(
                        type(exc).__name__, exc
                    )
                ) from exc
        else:
            stdout = stdout_bytes

        # We run subprocesses in foreground mode, so by the time Run's
        # constructor has returned, the subprocess is supposed to have
        # completed, and thus we are supposed to have an exit status code.
        assert subp.status is not None
        p = ProcessResult(subp.status, stdout)

        self.result.log += format_header("Status code", p.status)
        process_info["status"] = p.status
        process_info["output"] = Log(stdout)

        self.result.log += format_header(
            "Output",
            "\n"
            + truncated(str(process_info["output"]), truncate_logs_threshold),
        )

        # If requested, use its output for analysis
        if analyze_output:
            self.output += stdout

        if catch_error and p.status != 0:
            raise TestAbortWithFailure("non-zero status code")

        return p

    def add_test(self, dag: e3.collection.dag.DAG) -> None:
        self.add_fragment(dag, "run_wrapper")

    def push_success(self) -> None:
        """Set status to consider that the test passed."""
        # Given that we skip execution right after the test control evaluation,
        # there should be no way to call push_success in this case.
        assert not self.test_control.skip

        if self.test_control.xfail:
            self.result.set_status(TestStatus.XPASS, self.test_control.message)
        else:
            self.result.set_status(TestStatus.PASS)
        self.push_result()

    def push_skip(self, message: Optional[str]) -> None:
        """
        Consider that we skipped the test, set status accordingly.

        :param message: Label to explain the skipping.
        """
        self.result.set_status(TestStatus.SKIP, message)
        self.push_result()

    def push_error(self, message: Optional[str]) -> None:
        """
        Set status to consider that something went wrong during test execution.

        :param message: Message to explain what went wrong.
        """
        self.result.set_status(TestStatus.ERROR, message)
        self.push_result()

    def push_failure(self, message: Optional[str]) -> None:
        """
        Consider that the test failed and set status according to test control.

        :param message: Test failure description.
        """
        if self.test_control.xfail:
            status = TestStatus.XFAIL
            if self.test_control.message:
                message = "{} ({})".format(message, self.test_control.message)
        else:
            status = TestStatus.FAIL
        self.result.set_status(status, message)
        self.push_result()

    def set_up(self) -> None:
        """Run initialization operations before a test runs.

        Subclasses can override this to prepare testcase execution.

        Having a callback separate from "run" is useful when dealing with
        inheritance: overriding the "set_up" method in subclasses allows to
        append setup actions before the testcase execution actually takes place
        (in the "run" method).

        If everything happened in "run" method, that would not be possible
        unless re-implementing the "run" method in each subclass, with obvious
        code duplication issues.
        """
        pass

    def cleanup_working_dir(self) -> None:
        """Remove the working directory tree."""
        try:
            rm(self.working_dir(), True)
        except Exception:  # no cover
            # TODO (U222-013) For mysterious reasons, on Windows hosts,
            # sometimes executable files are still visible in the filesystem
            # even after the call to "os.unlink" returned with success. As a
            # result, removing the directory that contains them fails and thus
            # we get an exception.  At first we thought it could be related to
            # the system indexer
            # (https://superuser.com/questions/260375/why-would-system-continue-locking-
            # executable-file-handles-after-the-app-has-exit)
            # but this issue still occurs on systems that have it disabled.
            #
            # As far as we know (because we failed to pinpoint the exact reason
            # for this condition), these issues do not reveal any bug in tests
            # themselves, so silently ignore such errors.
            if self.env.host.os.name == "windows":
                self.result.log += (
                    f"\nError while cleaning up the working directory:"
                    f"\n{traceback.format_exc()}"
                    f"\nHost is running Windows: discarding this error..."
                )
            else:
                raise

    def tear_down(self) -> None:
        """Run finalization operations after a test has run.

        Subclasses can override this to run clean-ups after testcase execution.
        By default, this method removes the working directory (unless
        --disable-cleanup/--dev-temp is passed).

        See set_up's docstring for the rationale behind this API.
        """
        # Do nothing if cleanup is disabled, not requested for this result or
        # disabled for this driver.
        if (
            self.env.cleanup_mode == CleanupMode.NONE
            or (
                self.env.cleanup_mode == CleanupMode.PASSING
                and self.result.status
                in (
                    TestStatus.FAIL,
                    TestStatus.XFAIL,
                    TestStatus.XPASS,
                    TestStatus.ERROR,
                )
            )
            or not self.working_dir_cleanup_enabled
        ):
            return

        # If an error occurs during working dir cleanup, create a dedicated
        # error message and dump the content of the working directory tree
        # to help investigation.
        wd = self.working_dir()
        try:
            self.cleanup_working_dir()
        except Exception:
            result = TestResult(
                f"{self.test_name}__tear_down",
                env=self.test_env,
                status=TestStatus.ERROR,
            )

            result.log += (
                f"Error while removing the working directory {wd}:\n\n"
            )
            result.log += traceback.format_exc()
            result.log += "\nRemaining files:\n"
            for dirpath, _dirnames, filenames in os.walk(wd):
                if dirpath != wd:
                    result.log += f"  {os.path.relpath(dirpath, wd)}\n"
                for f in filenames:
                    fpath = os.path.join(dirpath, f)
                    result.log += f"  {os.path.relpath(fpath, wd)}\n"
            self.push_result(result)

    def run_wrapper(self, prev: Dict[str, Any], slot: int) -> None:
        # Make the slot (unique identifier for active jobs at a specific time)
        # available to the overridable methods.
        self.slot = slot

        # Make colors available for output if enabled testsuite-wide
        if self.env.enable_colors:  # interactive-only
            self.Fore = Fore
            self.Style = Style
        else:
            self.Fore = DummyColors()
            self.Style = DummyColors()

        # Create a test control for this test...
        try:
            self.test_control = self.test_control_creator.create(self)
        except ValueError as exc:
            return self.push_error(
                "Error while interpreting control: {}".format(exc)
            )

        # If test control tells us to skip the test, stop right here. Note that
        # if we have both skip and xfail, we are supposed not to execute the
        # test but still consider it as an expected failure (not just
        # "skipped").
        if self.test_control.skip:
            if self.test_control.xfail:
                return self.push_failure(self.test_control.message)
            else:
                return self.push_skip(self.test_control.message)

        # If requested, prepare the test working directory to initially be a
        # copy of the test directory.
        if self.copy_test_directory:
            sync_tree(
                self.test_env["test_dir"],
                self.test_env["working_dir"],
                delete=True,
            )

        # If the requested encoding is "binary", this actually means we will
        # handle binary data (i.e. no specific encoding). Create a binary log
        # accordingly.
        self.output = (
            Log(b"") if self.default_encoding == "binary" else Log("")
        )

        # Execute the subclass' "run" method and handle convenience test
        # aborting exception.
        try:
            self.set_up()
            self.run()
            self.analyze()
        except TestSkip as exc:
            return self.push_skip(str(exc))
        except TestAbortWithError as exc:
            return self.push_error(str(exc))
        except TestAbortWithFailure as exc:
            return self.push_failure(str(exc))
        finally:
            self.tear_down()

    def process_may_have_timed_out(self, result: ProcessResult) -> bool:
        """
        Return whether the process that yielded ``result`` may have timed out.

        This assumes that ``result`` is the returned value from a call to the
        ``shell`` method. Note that this uses simple heuristics to determine
        whether the process may have timed out, as this information is not
        preserved reliably under the hood: process is wrapped under the rlimit
        program, which just prints a known error message and returns some
        specific exit code in case of timeout.
        """
        if result.status != 2:
            return False

        if isinstance(result.out, str):
            return bool(TIMEOUT_OUTPUT_STR_RE.search(result.out))
        else:
            return bool(TIMEOUT_OUTPUT_BYTES_RE.search(result.out))

    def compute_failures(self) -> List[str]:
        """
        Analyze the testcase result and return the list of reasons for failure.

        This architecture allows to have multiple reasons for failures, for
        instance: unexpected computation result + presence of Valgrind
        diagnostics. The result is a list of short strings that describe the
        failures. This method is expected to write to ``self.result.log`` in
        order to convey more information if needed.

        By default, consider that the testcase succeeded if we reach the
        analysis step. Subclasses may override this to actually perform checks.
        """
        return []

    def analyze(self) -> None:
        """Analyze the testcase result, adjust status accordingly."""
        failures = self.compute_failures()
        if failures:
            self.push_failure(" | ".join(failures))
        else:
            self.push_success()



"""Tests for the "e3-find-skipped-tests" script."""

from e3.testsuite.find_skipped_tests import main
from e3.testsuite.result import TestStatus as Status

from .utils import create_report, create_result


def run(results_map, tmp_path, capsys):
    argv = []
    for dir_name, results in results_map.items():
        dir_path = tmp_path / dir_name
        dir_path.mkdir()
        create_report(results, dir_path)
        argv.append(str(dir_path))
    main(argv)
    captured = capsys.readouterr()
    assert not captured.err
    return captured.out


def test_no_skipped(tmp_path, capsys):
    assert (
        run(
            {
                "res1": [
                    create_result("foo", Status.SKIP),
                    create_result("bar", Status.PASS),
                ],
                "res2": [
                    create_result("foo", Status.XFAIL),
                    create_result("bar", Status.SKIP),
                ],
            },
            tmp_path,
            capsys,
        )
        == "All testcases are executed at least once.\n"
    )


def test_some_skipped(tmp_path, capsys):
    assert (
        run(
            {
                "res1": [
                    create_result("foo", Status.SKIP),
                    create_result("bar", Status.PASS),
                    create_result("baz", Status.SKIP),
                ],
                "res2": [
                    create_result("foo", Status.XFAIL),
                    create_result("bar", Status.SKIP),
                    create_result("qux", Status.SKIP),
                ],
            },
            tmp_path,
            capsys,
        )
        == "The following tests are never executed:\n  baz\n  qux\n"
    )
